import abc
import hashlib
import os
from abc import ABC
from typing import List, Union

import anndata
import numpy as np
import tensorflow as tf
from tqdm import tqdm

from sfaira.consts import AdataIdsSfaira, AdataIds
from sfaira.data.store.carts.single import CartSingle
from sfaira.data.store.stores.single import StoreSingleFeatureSpace
from sfaira.estimators.base import EstimatorBaseCelltype, EstimatorBaseEmbedding
from sfaira.models import BasicModelKeras
from sfaira.models.celltype import BasicModelKerasCelltype
from sfaira.models.embedding import BasicModelKerasEmbedding
from sfaira.versions.metadata import CelltypeUniverse, OntologyObo
from sfaira.versions.topologies import TopologyContainer
from .losses import LossLoglikelihoodNb, LossLoglikelihoodGaussian, LossCrossentropyAgg, KLLoss
from .metrics import custom_mse, custom_negll_nb, custom_negll_gaussian, \
    CustomAccAgg, CustomF1Classwise, CustomFprClasswise, CustomTprClasswise, custom_cce_agg


def prepare_sf(x):
    """
    Uses a minimal size factor of 1e-3 for total counts / 1e4
    """
    sf = np.asarray(x.sum(axis=1, keepdims=True))
    sf = np.log(np.maximum(sf / 1e4, 1e-3))
    return sf


def process_tf_dataset(dataset, mode: str, batch_size: int, cache: bool, shuffle_buffer_size: int, prefetch)\
        -> tf.data.Dataset:
    if cache:
        dataset = dataset.cache()
    if mode in ['train', 'train_val']:
        dataset = dataset.repeat()
        if shuffle_buffer_size > 0:
            # Only shuffle in train modes
            dataset = dataset.shuffle(buffer_size=shuffle_buffer_size, seed=None, reshuffle_each_iteration=True)
    dataset = dataset.batch(batch_size).prefetch(prefetch)
    return dataset


def get_optimizer(optimizer: str, lr: float):
    if optimizer.lower() == "adam":
        return tf.keras.optimizers.Adam(learning_rate=lr)
    elif optimizer.lower() == "sgd":
        return tf.keras.optimizers.SGD(learning_rate=lr)
    elif optimizer.lower() == "rmsprop":
        return tf.keras.optimizers.RMSprop(learning_rate=lr)
    elif optimizer.lower() == "adagrad":
        return tf.keras.optimizers.Adagrad(learning_rate=lr)
    else:
        assert False


def assemble_cbs(patience, lr_schedule_factor, lr_schedule_patience, lr_schedule_min_lr, lr_verbose, log_dir,
                 callbacks) -> List[tf.keras.callbacks.Callback]:
    cbs = [tf.keras.callbacks.TerminateOnNaN()]
    if patience is not None and patience > 0:
        cbs.append(tf.keras.callbacks.EarlyStopping(
            monitor='val_loss',
            patience=patience,
            restore_best_weights=True,
            verbose=1
        ))
    if lr_schedule_factor is not None and lr_schedule_factor < 1.:
        cbs.append(tf.keras.callbacks.ReduceLROnPlateau(
            monitor='val_loss',
            factor=lr_schedule_factor,
            patience=lr_schedule_patience,
            min_lr=lr_schedule_min_lr,
            verbose=lr_verbose
        ))
    if log_dir is not None:
        cbs.append(tf.keras.callbacks.TensorBoard(
            log_dir=log_dir,
            histogram_freq=0,
            batch_size=32,
            write_graph=False,
            write_grads=False,
            write_images=False,
            embeddings_freq=0,
            embeddings_layer_names=None,
            embeddings_metadata=None,
            embeddings_data=None,
            update_freq='epoch'
        ))
    if callbacks is not None:
        # callbacks needs to be a list
        cbs += callbacks
    return cbs


class EstimatorKeras(ABC):
    """
    Estimator base class for keras models.

    Important: Subclass implementing abstract classes also has to inherit from EstimatorBase class.
    """
    data: StoreSingleFeatureSpace
    model: Union[BasicModelKeras, None]
    weights: Union[np.ndarray, None]
    model_dir: Union[str, None]
    history: Union[dict, None]
    train_hyperparam: Union[dict, None]
    idx_train: Union[np.ndarray, None]
    idx_eval: Union[np.ndarray, None]
    idx_test: Union[np.ndarray, None]
    cache_path: str
    model_dir: Union[str, None]
    model_id: Union[str, None]
    md5: Union[str, None]

    @abc.abstractmethod
    def _get_cart(self, **kwargs) -> CartSingle:
        """
        Yield a generator based on which a tf dataset can be built.
        """
        pass

    @abc.abstractmethod
    def _tf_dataset_kwargs(self, mode: str):
        pass

    @abc.abstractmethod
    def split_train_val_test(self, val_split: float, test_split: Union[float, dict]):
        """
        Split indices in store into train, valiation and test split.
        """
        pass

    def get_one_time_tf_dataset(self, idx, mode, batch_size=None, prefetch=None):
        batch_size = 128 if batch_size is None else batch_size
        prefetch = 10 if prefetch is None else prefetch
        tf_kwargs = {
            "batch_size": batch_size,
            "cache": False,
            "prefetch": prefetch,
            "shuffle_buffer_size": 0,
        }
        train_gen = self._get_cart(idx=idx, mode=mode, retrieval_batch_size=128,
                                   randomized_batch_access=False)
        train_tf_dataset_kwargs = self._tf_dataset_kwargs(mode=mode)
        train_dataset = train_gen.adaptor(generator_type="tensorflow", **train_tf_dataset_kwargs)
        train_dataset = process_tf_dataset(dataset=train_dataset, mode=mode, **tf_kwargs)
        return train_dataset

    @abc.abstractmethod
    def _get_loss(self):
        pass

    @abc.abstractmethod
    def _metrics(self):
        pass

    def _compile_models(
            self,
            optimizer: tf.keras.optimizers.Optimizer
    ):
        self.model.training_model.compile(
            optimizer=optimizer,
            loss=self._get_loss(),
            metrics=self._metrics()
        )

    def train(
            self,
            optimizer: str,
            lr: float,
            epochs: int = 1000,
            max_steps_per_epoch: Union[int, None] = 20,
            batch_size: int = 128,
            validation_split: float = 0.1,
            test_split: Union[float, dict] = 0.,
            validation_batch_size: int = 256,
            max_validation_steps: Union[int, None] = 10,
            patience: int = 20,
            lr_schedule_min_lr: float = 1e-5,
            lr_schedule_factor: float = 0.2,
            lr_schedule_patience: int = 5,
            shuffle_buffer_size: Union[int, None] = None,
            cache_full: bool = False,
            randomized_batch_access: bool = True,
            retrieval_batch_size: int = 512,
            prefetch: Union[int, None] = 1,
            log_dir: Union[str, None] = None,
            callbacks: Union[list, None] = None,
            weighted: bool = False,
            verbose: int = 2
    ):
        """
        Train model.

        Uses validation loss and maximum number of epochs as termination criteria.

        :param optimizer: str corresponding to tf.keras optimizer to use for fitting.
        :param lr: Learning rate
        :param epochs: refer to tf.keras.models.Model.fit() documentation
        :param max_steps_per_epoch: Maximum steps per epoch.
        :param batch_size: refer to tf.keras.models.Model.fit() documentation
        :param validation_split: refer to tf.keras.models.Model.fit() documentation
            Refers to fraction of training data (1-test_split) to use for validation.
        :param test_split: Fraction of data to set apart for test model before train-validation split.
        :param validation_batch_size: Number of validation data observations to evaluate evaluation metrics on.
        :param max_validation_steps: Maximum number of validation steps to perform.
        :param patience: refer to tf.keras.models.Model.fit() documentation
        :param lr_schedule_min_lr: Minimum learning rate for learning rate reduction schedule.
        :param lr_schedule_factor: Factor to reduce learning rate by within learning rate reduction schedule
            when plateau is reached.
        :param lr_schedule_patience: Patience for learning rate reduction in learning rate reduction schedule.
        :param shuffle_buffer_size: tf.Dataset.shuffle(): buffer_size argument.
        :param cache_full: Whether to use tensorflow caching on full training and validation data.
        :param randomized_batch_access: Whether to randomize batches during reading (in generator). Lifts necessity of
            using a shuffle buffer on generator, however, batch composition stays unchanged over epochs unless there
            is overhangs in retrieval_batch_size in the raw data files, which often happens and results in modest
            changes in batch composition.
        :param log_dir: Directory to save tensorboard callback to. Disabled if None.
        :param callbacks: Add additional callbacks to the training call
        :param weighted:
        :param verbose:
        :return:
        """
        # Save training settings to allow model restoring.
        self.train_hyperparam = {"epochs": epochs, "max_steps_per_epoch": max_steps_per_epoch, "optimizer": optimizer,
                                 "lr": lr, "batch_size": batch_size, "validation_split": validation_split,
                                 "validation_batch_size": validation_batch_size,
                                 "max_validation_steps": max_validation_steps, "patience": patience,
                                 "lr_schedule_min_lr": lr_schedule_min_lr, "lr_schedule_factor": lr_schedule_factor,
                                 "lr_schedule_patience": lr_schedule_patience, "log_dir": log_dir, "weighted": weighted}

        # Set callbacks.
        cbs = assemble_cbs(patience=patience, lr_schedule_factor=lr_schedule_factor,
                           lr_schedule_patience=lr_schedule_patience, lr_schedule_min_lr=lr_schedule_min_lr,
                           lr_verbose=verbose, log_dir=log_dir, callbacks=callbacks)

        # Check randomisation settings:
        if shuffle_buffer_size is not None and shuffle_buffer_size > 0 and randomized_batch_access:
            raise ValueError("You are using shuffle_buffer_size and randomized_batch_access, this is likely not "
                             "intended.")
        shuffle_buffer_size = shuffle_buffer_size if shuffle_buffer_size is not None else 0
        if cache_full and randomized_batch_access:
            raise ValueError("You are using cache_full and randomized_batch_access, this is likely not intended.")
        self.split_train_val_test(val_split=validation_split, test_split=test_split)
        self._compile_models(optimizer=get_optimizer(optimizer=optimizer, lr=lr))

        tf_kwargs = {"batch_size": batch_size, "cache": cache_full, "prefetch": prefetch,
                     "shuffle_buffer_size": min(shuffle_buffer_size, len(self.idx_train))}
        train_gen = self._get_cart(idx=self.idx_train, mode='train', retrieval_batch_size=retrieval_batch_size,
                                   randomized_batch_access=randomized_batch_access, weighted=weighted)
        train_tf_dataset_kwargs = self._tf_dataset_kwargs(mode="train")
        train_dataset = train_gen.adaptor(generator_type="tensorflow", **train_tf_dataset_kwargs)
        train_dataset = process_tf_dataset(dataset=train_dataset, mode="train", **tf_kwargs)
        val_gen = self._get_cart(idx=self.idx_train, mode='train', retrieval_batch_size=retrieval_batch_size,
                                 randomized_batch_access=randomized_batch_access, weighted=weighted)
        val_tf_dataset_kwargs = self._tf_dataset_kwargs(mode="train_val")
        val_dataset = val_gen.adaptor(generator_type="tensorflow", **val_tf_dataset_kwargs)
        val_dataset = process_tf_dataset(dataset=val_dataset, mode="train", **tf_kwargs)

        steps_per_epoch = min(max(len(self.idx_train) // batch_size, 1), max_steps_per_epoch)
        validation_steps = min(max(len(self.idx_eval) // validation_batch_size, 1), max_validation_steps)

        self.history = self.model.training_model.fit(
            x=train_dataset,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            callbacks=cbs,
            validation_data=val_dataset,
            validation_steps=validation_steps,
            verbose=verbose
        ).history

    def load_pretrained_weights(self):
        """
        Loads model weights from local directory or zenodo.
        """
        if self.model_dir.startswith('http'):
            # Remote repo
            if not os.path.exists(self.cache_path):
                os.makedirs(self.cache_path)

            import urllib.request
            from urllib.parse import urljoin
            from urllib.error import HTTPError
            try:
                urllib.request.urlretrieve(self.model_dir,
                                           os.path.join(self.cache_path, os.path.basename(self.model_dir)))
                fn = os.path.join(self.cache_path, os.path.basename(self.model_dir))
            except HTTPError:
                try:
                    urllib.request.urlretrieve(urljoin(self.model_dir, f'{self.model_id}_weights.h5'),
                                               os.path.join(self.cache_path, f'{self.model_id}_weights.h5'))
                    fn = os.path.join(self.cache_path, f"{self.model_id}_weights.h5")
                except HTTPError:
                    try:
                        urllib.request.urlretrieve(
                            urljoin(self.model_dir, f'{self.model_id}_weights.data-00000-of-00001'),
                            os.path.join(self.cache_path, f'{self.model_id}_weights.data-00000-of-00001')
                        )
                        fn = os.path.join(self.cache_path, f"{self.model_id}_weights.data-00000-of-00001")
                    except HTTPError:
                        raise FileNotFoundError('cannot find remote weightsfile')
        else:
            # Local repo
            if not self.model_dir:
                raise ValueError('the model_id is set but the path to the model is empty')
            if os.path.isfile(self.model_dir) \
                    and not self.model_dir.endswith(".h5") \
                    and not self.model_dir.endswith(".data-00000-of-00001"):
                raise ValueError('weights files saved in h5 format need to have an h5 file extension')

            if os.path.isfile(self.model_dir):
                fn = self.model_dir
            elif os.path.isfile(os.path.join(self.model_dir, f"{self.model_id}_weights.data-00000-of-00001")):
                fn = os.path.join(self.model_dir, f"{self.model_id}_weights.data-00000-of-00001")
            elif os.path.isfile(os.path.join(self.model_dir, f"{self.model_id}_weights.h5")):
                fn = os.path.join(self.model_dir, f"{self.model_id}_weights.h5")
            else:
                raise ValueError('the weightsfile could not be found')

        if self.md5 is not None:
            self._assert_md5_sum(fn, self.md5)
        if fn.endswith(".data-00000-of-00001"):
            self.model.training_model.load_weights(".".join(fn.split(".")[:-1]))
        else:
            self.model.training_model.load_weights(fn)

    def save_weights_to_cache(self):
        if not os.path.exists(os.path.join(self.cache_path, 'weights')):
            os.makedirs(os.path.join(self.cache_path, 'weights'))
        fn = os.path.join(self.cache_path, 'weights', f"{self.model_id}_weights_cache.h5")
        self.model.training_model.save_weights(fn)

    def load_weights_from_cache(self):
        fn = os.path.join(self.cache_path, 'weights', f"{self.model_id}_weights_cache.h5")
        self.model.training_model.load_weights(fn)

    @staticmethod
    def _assert_md5_sum(
            fn,
            target_md5
    ):
        with open(fn, 'rb') as f:
            hsh = hashlib.md5(f.read()).hexdigest()
        if not hsh == target_md5:
            raise ValueError("md5 of %s did not match expectation" % fn)

    def init_model(self, clear_weight_cache=True, override_hyperpar=None):
        """
        Instantiate the model.
        :return:
        """
        if clear_weight_cache:
            if os.path.exists(os.path.join(self.cache_path, 'weights')):
                for file in os.listdir(os.path.join(self.cache_path, 'weights')):
                    file_path = os.path.join(os.path.join(self.cache_path, 'weights'), file)
                    os.remove(file_path)


class EstimatorKerasEmbedding(EstimatorBaseEmbedding, EstimatorKeras):
    """
    Estimator class for the embedding model.
    """

    model: Union[BasicModelKerasEmbedding, None]

    def __init__(
            self,
            data: Union[anndata.AnnData, np.ndarray, StoreSingleFeatureSpace],
            model_dir: Union[str, None],
            model_id: Union[str, None],
            model_topology: TopologyContainer,
            weights_md5: Union[str, None] = None,
            cache_path: str = os.path.join('cache', ''),
            adata_ids: AdataIds = AdataIdsSfaira()
    ):
        super(EstimatorKerasEmbedding, self).__init__(
            data=data,
            model_dir=model_dir,
            model_class="embedding",
            model_id=model_id,
            model_topology=model_topology,
            weights_md5=weights_md5,
            cache_path=cache_path,
            adata_ids=adata_ids
        )

    def init_model(
            self,
            clear_weight_cache: bool = True,
            override_hyperpar: Union[dict, None] = None
    ):
        """
        instantiate the model
        :return:
        """
        super().init_model(clear_weight_cache=clear_weight_cache)
        if self.model_type == 'vae':
            from sfaira.models.embedding import ModelVaeVersioned as Model
        elif self.model_type == 'ae':
            from sfaira.models.embedding import ModelAeVersioned as Model
        elif self.model_type == 'linear' or self.model_type == 'nmf':
            from sfaira.models.embedding import ModelLinearVersioned as Model
        elif self.model_type == 'vaeiaf':
            from sfaira.models.embedding import ModelVaeIAFVersioned as Model
        elif self.model_type == 'vaevamp':
            from sfaira.models.embedding import ModelVaeVampVersioned as Model
        else:
            raise ValueError(f'unknown model type {self.model_type} for EstimatorKerasEmbedding')
        self.model = Model(
            topology_container=self.topology_container,
            override_hyperpar=override_hyperpar
        )

    def _tf_dataset_kwargs(self, mode: str):
        # Determine model type [ae, vae(iaf, vamp)]
        model_type = "vae" if self.model_type[:3] == "vae" else "ae"
        output_types_x = (tf.float32, tf.float32)
        output_shapes_x = (self.data.n_vars, 1)
        if mode == 'predict':  # Output shape is same for predict mode regardless of model type
            output_types = output_types_x,
            output_shapes = output_shapes_x,
        elif model_type == "vae":
            output_types = (output_types_x, (tf.float32, tf.float32))
            output_shapes = (output_shapes_x, (self.data.n_vars, 1))
        else:
            output_types = (output_types_x, (tf.float32, ))
            output_shapes = (output_shapes_x, (self.data.n_vars, ))
        return {"output_types": output_types, "output_shapes": output_shapes}

    def _get_cart(
            self,
            idx: Union[np.ndarray, None],
            mode: str,
            retrieval_batch_size: int,
            randomized_batch_access: bool,
            **kwargs
    ):
        # Define constants used by map_fn in outer name space so that they are not created for each sample.
        model_type = "vae" if self.model_type[:3] == "vae" else "ae"

        def map_fn(x_sample, obs_sample):
            x_sample = np.asarray(x_sample)
            sf_sample = prepare_sf(x=x_sample)
            output_x = (x_sample, sf_sample)
            if mode == 'predict':
                output = output_x,
            elif model_type == "vae":
                output = output_x, (x_sample, sf_sample)
            else:
                output = output_x, (x_sample, )
            return output

        g = self.data.checkout(idx=idx, retrieval_batch_size=retrieval_batch_size, obs_keys=[], map_fn=map_fn,
                               return_dense=True, randomized_batch_access=randomized_batch_access,
                               random_access=False)
        return g

    def _get_loss(self):
        if self.topology_container.topology["hyper_parameters"]["output_layer"] in [
            "nb", "nb_const_disp", "nb_shared_disp"
        ]:
            reconstruction_loss_vae = LossLoglikelihoodNb(average=False)  # TODO maybe handly kwargs via functional?
            reconstruction_loss = LossLoglikelihoodNb(average=True)
        elif self.topology_container.topology["hyper_parameters"]["output_layer"] in [
            "gaussian", "gaussian_const_disp", "gaussian_shared_disp"
        ]:
            reconstruction_loss_vae = LossLoglikelihoodGaussian(average=False)
            reconstruction_loss = LossLoglikelihoodGaussian(average=True)
        else:
            raise ValueError(self.topology_container.topology["hyper_parameters"]["output_layer"])

        if self.model_type[:3] == "vae":  # TODO too hacky
            return {
                "neg_ll": reconstruction_loss_vae,
                "kl": KLLoss()
            }
        else:
            return {"neg_ll": reconstruction_loss}

    def _metrics(self):
        if self.topology_container.topology["hyper_parameters"]["output_layer"] in [
            "nb", "nb_const_disp", "nb_shared_disp"
        ]:
            custom_negll = custom_negll_nb
        elif self.topology_container.topology["hyper_parameters"]["output_layer"] in [
            "gaussian", "gaussian_const_disp", "gaussian_shared_disp"
        ]:
            custom_negll = custom_negll_gaussian
        else:
            raise ValueError(self.topology_container.topology["hyper_parameters"]["output_layer"])

        custom_negll.__name__ = "custom_negll"

        return {"neg_ll": [custom_mse, custom_negll]}

    def evaluate_any(self, idx, batch_size: int = 128, max_steps: int = np.inf):
        """
        Evaluate the custom model on any local data.

        Defaults to run on full data if idx is None.

        :param idx: Indices of observations to evaluate on. Evaluates on all observations if None.
        :param batch_size: Batch size for evaluation.
        :param max_steps: Maximum steps before evaluation round is considered complete.
        :return: Dictionary of metric names and values.
        """
        idx = self._process_idx_for_eval(idx=idx)
        if idx is not None:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='eval')
            steps = min(max(len(idx) // batch_size, 1), max_steps)
            results = self.model.training_model.evaluate(x=dataset, steps=steps)
            return dict(zip(self.model.training_model.metrics_names, results))
        else:
            return {}

    def evaluate(self, batch_size: int = 128, max_steps: int = np.inf):
        """
        Evaluate the custom model on test data.

        Defaults to run on full data if idx_test was not set before, ie. train() has not been called before.

        :param batch_size: Batch size for evaluation.
        :param max_steps: Maximum steps before evaluation round is considered complete.
        :return: Dictionary of metric names and values.
        """
        idx = self._process_idx_for_eval(idx=self.idx_test)
        if idx is not None:
            return self.evaluate_any(idx=self.idx_test, batch_size=batch_size, max_steps=max_steps)
        else:
            return {}

    def predict(self, batch_size: int = 128):
        """
        return the prediction of the model

        :return:
        prediction
        """
        idx = self._process_idx_for_eval(idx=self.idx_test)
        if idx is not None:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='predict')
            return self.model.predict_reconstructed(x=dataset)
        else:
            return np.array([])

    def predict_embedding(self, batch_size: int = 128, variational: bool = False):
        """
        return the prediction in the latent space (z_mean for variational models)

        :params variational: Whether toreturn the prediction of z, z_mean, z_log_var in the variational latent space.
        :return:
        latent space
        """
        idx = self._process_idx_for_eval(idx=self.idx_test)
        if len(idx) > 0:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='predict')
            return self.model.predict_embedding(x=dataset, variational=variational)
        else:
            return np.array([])

    def compute_gradients_input(
            self,
            batch_size: int = 128,
            test_data: bool = False,
            abs_gradients: bool = True,
            per_celltype: bool = False
    ):
        # TODO may need to be adapted to new dataset / generator format
        if test_data:
            idx = self.idx_test
            if self.idx_test is None:
                num_samples = 10000
                idx = np.random.randint(0, self.data.n_obs, num_samples)
            n_obs = len(idx)
        else:
            idx = None
            n_obs = self.data.n_obs

        ds = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='gradient_method')

        if per_celltype:
            cell_to_id = self._get_class_dict(obs_key=self._adata_ids.cell_type)
            cell_names = cell_to_id.keys()
            cell_id = cell_to_id.values()
            id_to_cell = dict([(key, value) for (key, value) in zip(cell_id, cell_names)])
            grads_x = dict([(key, 0) for key in cell_names])
            counts = dict([(key, 0) for key in cell_names])
        else:
            grads_x = 0
        # Loop over sub-selected data set and sum gradients across all selected observations.
        # TODO: fix bug for vaeiaf model. This function can not be called for vaeiaf model
        if self.model_type[:3] == "vae":
            model = tf.keras.Model(
                self.model.training_model.input,
                self.model.encoder_model.output[0]
            )
            latent_dim = self.model.encoder_model.output[0].shape[1]
            input_dim = self.model.training_model.input[0].shape[1]
        else:
            model = tf.keras.Model(
                self.model.training_model.input,
                self.model.encoder_model.output
            )
            latent_dim = self.model.encoder_model.output[0].shape[0]
            input_dim = self.model.training_model.input[0].shape[1]

        @tf.function
        def get_gradients(x_batch):
            x, sf = x_batch
            with tf.GradientTape(persistent=True) as tape:
                tape.watch(x)
                model_out = model((x, sf))
            if abs_gradients:
                def f(xx):
                    return abs(xx)
            else:
                def f(xx):
                    return xx
            # marginalize on batch level and then accumulate batches
            # batch_jacobian gives output of size: (batch_size, latent_dim, input_dim)
            return f(tape.batch_jacobian(model_out, x))

        for step, (x_batch, y_batch) in tqdm(enumerate(ds), total=np.ceil(n_obs / batch_size)):
            batch_gradients = get_gradients(x_batch).numpy()
            _, y = y_batch
            if per_celltype:
                for i in np.unique(y):
                    grads_x[id_to_cell[i]] += np.sum(batch_gradients[y == i, :, :], axis=0)
                    counts[id_to_cell[i]] += np.sum(y == i)
            else:
                grads_x += np.sum(batch_gradients, axis=0)
        if per_celltype:
            for cell in cell_names:
                print(f'{cell} with {counts[cell]} observations')
                grads_x[cell] = grads_x[cell] / counts[cell] if counts[cell] > 0 else np.zeros((latent_dim, input_dim))

            return {'gradients': grads_x, 'counts': counts}
        else:
            return grads_x / n_obs


class EstimatorKerasCelltype(EstimatorBaseCelltype, EstimatorKeras):
    """
    Estimator class for the cell type model.
    """

    celltype_universe: CelltypeUniverse
    model: Union[BasicModelKerasCelltype, None]

    def __init__(
            self,
            data: Union[anndata.AnnData, StoreSingleFeatureSpace],
            model_dir: Union[str, None],
            model_id: Union[str, None],
            model_topology: TopologyContainer,
            weights_md5: Union[str, None] = None,
            cache_path: str = os.path.join('cache', ''),
            celltype_ontology: Union[OntologyObo, None] = None,
            max_class_weight: float = 1e3,
            remove_unlabeled_cells: bool = True,
            adata_ids: AdataIds = AdataIdsSfaira()
    ):
        super(EstimatorKerasCelltype, self).__init__(
            data=data,
            model_dir=model_dir,
            model_id=model_id,
            model_topology=model_topology,
            weights_md5=weights_md5,
            cache_path=cache_path,
            celltype_ontology=celltype_ontology,
            max_class_weight=max_class_weight,
            remove_unlabeled_cells=remove_unlabeled_cells,
            adata_ids=adata_ids
        )

    def init_model(
            self,
            clear_weight_cache: bool = True,
            override_hyperpar: Union[dict, None] = None
    ):
        """
        instantiate the model
        :return:
        """
        super().init_model(clear_weight_cache=clear_weight_cache)
        if self.model_type == "marker":
            from sfaira.models.celltype import CellTypeMarkerVersioned as Model
        elif self.model_type == "mlp":
            from sfaira.models.celltype import CellTypeMlpVersioned as Model
        else:
            raise ValueError('unknown topology %s for EstimatorKerasCelltype' % self.model_type)

        self.model = Model(
            celltypes_version=self.celltype_universe,
            topology_container=self.topology_container,
            override_hyperpar=override_hyperpar
        )

    def _tf_dataset_kwargs(self, mode):
        output_types_x = (tf.float32,)
        output_shapes_x = (tf.TensorShape([self.data.n_vars]), )
        if mode == 'predict':
            output_types = (output_types_x, )
            output_shapes = (output_shapes_x, )
        else:
            output_types = (output_types_x, tf.float32)
            output_shapes = (output_shapes_x, tf.TensorShape([self.ntypes]))
        return {"output_types": output_types, "output_shapes": output_shapes}

    def _get_cart(
            self,
            idx: Union[np.ndarray, None],
            mode: str,
            retrieval_batch_size: int,
            randomized_batch_access: bool,
            weighted: bool = False,
            **kwargs
    ) -> CartSingle:
        # Define constants used by map_fn in outer name space so that they are not created for each sample.
        if weighted:
            raise ValueError("using weights with store is not supported yet")
        yield_labels = mode in ["train", "train_val", "eval", "test"]
        if yield_labels:
            onehot_encoder = self._one_hot_encoder()

        def map_fn(x_sample, obs_sample):
            x_sample = np.asarray(x_sample)
            output_x = (x_sample, )
            if yield_labels:
                y_sample = onehot_encoder(obs_sample[self._adata_ids.cell_type + self._adata_ids.onto_id_suffix].values)
                # Only yield observations with valid label:
                idx_keep = y_sample.sum(axis=1) > 0.
                if not np.all(idx_keep):
                    idx_keep = np.where(idx_keep)[0]
                    output_x = tuple([x[idx_keep, :] for x in output_x])
                    y_sample = y_sample[idx_keep, :]
                output_y = y_sample
                output = output_x, output_y
            else:
                output = output_x,
            return output

        g = self.data.checkout(idx=idx, retrieval_batch_size=retrieval_batch_size,
                               obs_keys=[self._adata_ids.cell_type + self._adata_ids.onto_id_suffix], map_fn=map_fn,
                               return_dense=True, randomized_batch_access=randomized_batch_access,
                               random_access=False)
        return g

    def _get_loss(self):
        return LossCrossentropyAgg()

    def _metrics(self):
        return [
            "accuracy",
            custom_cce_agg,
            CustomAccAgg(),
            CustomF1Classwise(k=self.ntypes),
            CustomFprClasswise(k=self.ntypes),
            CustomTprClasswise(k=self.ntypes)
        ]

    def predict(self, batch_size: int = 128, max_steps: int = np.inf):
        """
        Return the prediction of the model

        :param batch_size: Batch size for evaluation.
        :param max_steps: Maximum steps before evaluation round is considered complete.
        :return: Prediction tensor.
        """
        idx = self._process_idx_for_eval(idx=self.idx_test)
        if len(idx) > 0:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='predict')
            return self.model.training_model.predict(x=dataset)
        else:
            return np.array([])

    def ytrue(self, batch_size: int = 128, max_steps: int = np.inf):
        """
        Return the true labels of the test set.

        :return: true labels
        """
        idx = self._process_idx_for_eval(idx=self.idx_test)
        if len(idx) > 0:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='eval')
            y_true = []
            for _, y in dataset.as_numpy_iterator():
                y_true.append(y)
            y_true = np.concatenate(y_true, axis=0)
            return y_true
        else:
            return np.array([])

    def evaluate_any(self, idx, batch_size: int = 128, max_steps: int = np.inf, weighted: bool = False):
        """
        Evaluate the custom model on any local data.

        Defaults to run on full data if idx is None.

        :param idx: Indices of observations to evaluate on. Evaluates on all observations if None.
        :param batch_size: Batch size for evaluation.
        :param max_steps: Maximum steps before evaluation round is considered complete.
        :param weighted: Whether to use class weights in evaluation.
        :return: Dictionary of metric names and values.
        """
        idx = self._process_idx_for_eval(idx=idx)
        if len(idx) > 0:
            dataset = self.get_one_time_tf_dataset(idx=idx, batch_size=batch_size, mode='eval')
            results = self.model.training_model.evaluate(x=dataset)
            return dict(zip(self.model.training_model.metrics_names, results))
        else:
            return {}

    def evaluate(self, batch_size: int = 128, max_steps: int = np.inf, weighted: bool = False):
        """
        Evaluate the custom model on local data.

        Defaults to run on full data if idx_test was not set before, ie. train() has not been called before.

        :param batch_size: Batch size for evaluation.
        :param max_steps: Maximum steps before evaluation round is considered complete.
        :param weighted: Whether to use class weights in evaluation.
        :return: Dictionary of metric names and values.
        """
        return self.evaluate_any(idx=self.idx_test, batch_size=batch_size, max_steps=max_steps, weighted=weighted)

    def compute_gradients_input(
            self,
            test_data: bool = False,
            abs_gradients: bool = True
    ):
        # TODO may need to be adapted to new dataset / generator format
        if test_data:
            idx = self.idx_test
            n_obs = len(self.idx_test)
        else:
            idx = None
            n_obs = self.data.n_obs

        # to get a tf.GradientTape compatible data set
        ds = self.get_one_time_tf_dataset(idx=idx, batch_size=64, mode='train_val')
        grads_x = 0
        # Loop over sub-selected data set and sum gradients across all selected observations.
        model = tf.keras.Model(
            self.model.training_model.input,
            self.model.training_model.output
        )

        for step, (x_batch, _, _) in enumerate(ds):
            print("compute gradients wrt. input: batch %i / %i." % (step + 1, np.ceil(n_obs / 64)))
            x = x_batch
            with tf.GradientTape(persistent=True) as tape:
                tape.watch(x)
                model_out = model(x)
            if abs_gradients:
                def f(x):
                    return abs(x)
            else:
                def f(x):
                    return x
            # marginalize on batch level and then accumulate batches
            # batch_jacobian gives output of size: (batch_size, latent_dim, input_dim)
            batch_gradients = f(tape.batch_jacobian(model_out, x).numpy())
            grads_x += np.sum(batch_gradients, axis=0)
        return grads_x



from typing import List, Union
from sfaira.versions.metadata import Ontology

UNS_STRING_META_IN_OBS = "__obs__"


def is_child(
        query,
        ontology: Union[Ontology, bool, int, float, str, List[bool], List[int], List[float], List[str]],
        ontology_parent=None,
) -> True:
    """
    Check whether value is from set of allowed values using ontology.

    :param query: Value to attempt to set, only yield a single value and not a list.
    :param ontology: Constraint for values.
        Either ontology instance used to constrain entries, or list of allowed values.
    :param ontology_parent: If ontology is a DAG, not only check if node is a DAG node but also whether it is a child
        of this parent node.
    :return: Whether attempted term is sub-term of allowed term in ontology
    """
    if ontology_parent is None and ontology is None:
        return True
    else:
        if isinstance(ontology, Ontology):
            if ontology_parent is None:
                return ontology.is_node(query)
            else:
                return ontology.is_a(query=query, reference=ontology_parent)
        elif ontology is None:
            return query == ontology_parent
        else:
            raise ValueError(f"did not recognize ontology type {type(ontology)}")


def clean_string(s):
    if s is not None:
        s = s.replace(',', '').replace(' ', '').replace('-', '').replace('_', '').replace("'", '').lower()
    return s


def get_directory_formatted_doi(x: str) -> str:
    return "d" + "_".join("_".join("_".join(x.split("/")).split(".")).split("-"))


def identify_tsv(fn: str, ontology_names: List[str]) -> str:
    """
    Process ontology:
    1. In older data loader versions (<1.1), there is only one tsv file without suffix which
    corresponds to the cell type ontology.
    TODO: This could be deprecated in the future if these data loaders are updated.
    2. In versions >=1.1, the suffix of the tsv is the ontology.
    """
    y = [x for x in ontology_names if fn.split(".tsv")[0].endswith(x)]
    if len(y) == 0:  # <v1.1
        y = "cell_type"
    else:  # >=v1.1
        y = y[0]
    return y



"""
Settings class which for example holds paths to cache directories used throughout the code.
"""

import os


SFAIRA_REPO_URL = "https://zenodo.org/record/4836517/files/"


class SfairaConfig:
    """\
    Config manager for sfaira.
    """

    def __init__(self):
        self.sfaira_repo_url = SFAIRA_REPO_URL
        self._cachedir_base = os.path.join(os.path.expanduser("~"), ".cache", "sfaira")
        self._cachedir_databases = os.path.join(self._cachedir_base, "dataset_meta")
        self._cachedir_databases_cellxgene = os.path.join(self._cachedir_databases, "cellxgene")
        self._cachedir_genomes = os.path.join(self._cachedir_base, "genomes")
        self._cachedir_ontologies = os.path.join(self._cachedir_base, "ontologies")

    @property
    def cachedir_base(self) -> str:
        os.makedirs(self._cachedir_base, exist_ok=True)
        return self._cachedir_base

    @cachedir_base.setter
    def cachedir_base(self, cachedir_base):
        if not isinstance(cachedir_base, str):
            raise ValueError(f"cachedir_base needs to be provided as a string, was {type(cachedir_base)}")
        if cachedir_base == "repo":
            cachedir_base = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "cache")
        self._cachedir_base = cachedir_base

    @property
    def cachedir_databases(self) -> str:
        os.makedirs(self._cachedir_databases, exist_ok=True)
        return self._cachedir_databases

    @cachedir_databases.setter
    def cachedir_databases(self, cachedir_databases):
        raise ValueError("cachedir_databases cannot be set manually as it is defined as a subdirectory of"
                         " cachedir_base. please modify cachedir_base instead")

    @property
    def cachedir_databases_cellxgene(self) -> str:
        os.makedirs(self._cachedir_databases_cellxgene, exist_ok=True)
        return self._cachedir_databases_cellxgene

    @cachedir_databases_cellxgene.setter
    def cachedir_databases_cellxgene(self, cachedir_databases_cellxgene):
        raise ValueError("cachedir_databases_cellxgene cannot be set manually as it is defined as a subdirectory"
                         " of cachedir_base. please modify cachedir_base instead")

    @property
    def cachedir_genomes(self) -> str:
        os.makedirs(self._cachedir_genomes, exist_ok=True)
        return self._cachedir_genomes

    @cachedir_genomes.setter
    def cachedir_genomes(self, cachedir_genomes):
        raise ValueError("cachedir_genomes cannot be set manually as it is defined as a subdirectory of cachedir_base."
                         "please modify cachedir_base instead")

    @property
    def cachedir_ontologies(self) -> str:
        os.makedirs(self._cachedir_ontologies, exist_ok=True)
        return self._cachedir_ontologies

    @cachedir_ontologies.setter
    def cachedir_ontologies(self, cachedir_ontologies):
        raise ValueError("cachedir_ontologies cannot be set manually as it is defined as a subdirectory of cachedir_base. please modify cachedir_base instead")


settings = SfairaConfig()



from __future__ import annotations

import abc
import anndata
from anndata.utils import make_index_unique
import numpy as np
import pandas as pd
import os
from os import PathLike
import pandas
import scipy.sparse
from typing import Dict, List, Tuple, Union
import warnings
import urllib.request
import urllib.parse
import urllib.error
import cgi
import ssl

from sfaira.versions.genomes import GenomeContainer
from sfaira.versions.metadata import Ontology, OntologyHierarchical, CelltypeUniverse
from sfaira.consts import AdataIds, AdataIdsCellxgene_v2_0_0, AdataIdsSfaira, META_DATA_FIELDS, OCS
from sfaira.data.dataloaders.base.utils import identify_tsv
from sfaira.data.dataloaders.export_adaptors import cellxgene_export_adaptor
from sfaira.data.store.io.io_dao import write_dao
from sfaira.data.dataloaders.base.utils import is_child, get_directory_formatted_doi
from sfaira.data.utils import collapse_matrix, read_yaml, subset_adata_genes
from sfaira.consts.utils import clean_id_str
from sfaira.versions.metadata.maps import prepare_ontology_map_tab


load_doc = \
    """
    :param remove_gene_version: Remove gene version string from ENSEMBL ID so that different versions in different data sets are superimposed.
    :param match_to_reference: Reference genomes name or False to keep original feature space.
    :param load_raw: Loads unprocessed version of data if available in data loader.
    :param allow_caching: Whether to allow method to cache adata object for faster re-loading.
    """


class DatasetBase(abc.ABC):
    adata: Union[None, anndata.AnnData]
    _meta: Union[None, pandas.DataFrame]
    data_dir_base: Union[None, str]
    meta_path: Union[None, str]
    cache_path: Union[None, str]
    id: Union[None, str]
    genome: Union[None, str]
    supplier: str

    layer_counts: Union[None, str]
    layer_processed: Union[None, str]
    layer_spliced_counts: Union[None, str]
    layer_spliced_processed: Union[None, str]
    layer_unspliced_counts: Union[None, str]
    layer_unspliced_processed: Union[None, str]
    layer_velocity: Union[None, str]

    _assay_sc: Union[None, str]
    _assay_differentiation: Union[None, str]
    _assay_type_differentiation: Union[None, str]
    _author: Union[None, str]
    _bio_sample: Union[None, str]
    _cell_line: Union[None, str]
    _cell_type: Union[None, str]
    _default_embedding: Union[None, str]
    _development_stage: Union[None, str]
    _disease: Union[None, str]
    _doi_journal: Union[None, str]
    _doi_preprint: Union[None, str]
    _download_url_data: Union[Tuple[List[None]], Tuple[List[str]], None]
    _download_url_meta: Union[Tuple[List[None]], Tuple[List[str]], None]
    _ethnicity: Union[None, str]
    _feature_type: Union[None, str]
    _feature_reference: Union[None, str]
    _id: Union[None, str]
    _individual: Union[None, str]
    _organ: Union[None, str]
    _organism: Union[None, str]
    _primary_data: Union[None, bool]
    _sex: Union[None, str]
    _source: Union[None, str]
    _sample_source: Union[None, str]
    _state_exact: Union[None, str]
    _title: Union[None, str]
    _year: Union[None, int]
    gm: Union[None, str]
    treatment: Union[None, str]

    _bio_sample_obs_key: Union[None, str]
    _tech_sample_obs_key: Union[None, str]
    assay_sc_obs_key: Union[None, str]
    assay_differentiation_obs_key: Union[None, str]
    assay_type_differentiation_obs_key: Union[None, str]
    cell_type_obs_key: Union[None, str]
    development_stage_obs_key: Union[None, str]
    disease_obs_key: Union[None, str]
    ethnicity_obs_key: Union[None, str]
    gm_obs_key: Union[None, str]
    individual_obs_key: Union[None, str]
    organ_obs_key: Union[None, str]
    organism_obs_key: Union[None, str]
    sample_source_obs_key: Union[None, str]
    sex_obs_key: Union[None, str]
    state_exact_obs_key: Union[None, str]
    treatment_obs_key: Union[None, str]

    _feature_reference: Union[None, str]
    _feature_type: Union[None, str]

    feature_id_var_key: Union[None, str]
    feature_reference_var_key: Union[None, str]
    feature_symbol_var_key: Union[None, str]
    feature_type_var_key: Union[None, str]

    spatial_x_coord_obs_key: Union[None, str]
    spatial_y_coord_obs_key: Union[None, str]
    spatial_z_coord_obs_key: Union[None, str]
    vdj_vj_1_obs_key_prefix: Union[None, str]
    vdj_vj_2_obs_key_prefix: Union[None, str]
    vdj_vdj_1_obs_key_prefix: Union[None, str]
    vdj_vdj_2_obs_key_prefix: Union[None, str]
    vdj_c_call_obs_key_suffix: Union[None, str]
    vdj_consensus_count_obs_key_suffix: Union[None, str]
    vdj_d_call_obs_key_suffix: Union[None, str]
    vdj_duplicate_count_obs_key_suffix: Union[None, str]
    vdj_j_call_obs_key_suffix: Union[None, str]
    vdj_junction_obs_key_suffix: Union[None, str]
    vdj_junction_aa_obs_key_suffix: Union[None, str]
    vdj_locus_obs_key_suffix: Union[None, str]
    vdj_productive_obs_key_suffix: Union[None, str]
    vdj_v_call_obs_key_suffix: Union[None, str]

    _celltype_universe: Union[None, CelltypeUniverse]
    _ontology_class_maps: Union[dict]

    load_raw: Union[None, bool]
    mapped_features: Union[None, str, bool]
    remove_gene_version: Union[None, bool]
    subset_gene_type: Union[None, str]
    streamlined_meta: bool

    sample_fn: Union[None, str]
    _sample_fns: Union[None, List[str]]

    _additional_annotation_key: Union[None, str]

    def __init__(
            self,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            load_func=None,
            dict_load_func_annotation=None,
            yaml_path: Union[str, None] = None,
            sample_fn: Union[str, None] = None,
            sample_fns: Union[List[str], None] = None,
            additional_annotation_key: Union[str, None] = None,
            **kwargs
    ):
        """

        :param data_path:
        :param meta_path:
        :param cache_path:
        :param load_func: Function to load data from disk into memory.

            Signature: load(data_dir, sample_fn, **kwargs)
        :param dict_load_func_annotation: Dictionary of functions to load additional observatino-wise annotation. The
            functions in the values of the dictionary can be selected via  self.additional_annotation_key which needs
            to correspond to a key of the dictionary.

            Signature: Dict[str, load_annotation(data_dir, sample_fn, additional_annotation_key, **kwargs)]
        :param yaml_path:
        :param sample_fn:
        :param sample_fns:
        :param additional_annotation_key: Key used by dict_load_func_annotation to identify which additional annotation
            is to be loaded.
        :param kwargs:
        """
        self._adata_ids = AdataIdsSfaira()
        self.ontology_container_sfaira = OCS  # Using a pre-instantiated version of this yields drastic speed-ups.

        self.adata = None
        self.meta = None
        self.genome = None
        self.data_dir_base = data_path
        self.meta_path = meta_path
        self.cache_path = cache_path

        self.layer_counts = None
        self.layer_processed = None
        self.layer_spliced_counts = None
        self.layer_spliced_processed = None
        self.layer_unspliced_counts = None
        self.layer_unspliced_processed = None
        self.layer_velocity = None

        self._author = None
        self._assay_sc = None
        self._assay_differentiation = None
        self._assay_type_differentiation = None
        self._bio_sample = None
        self._cell_line = None
        self._cell_type = None
        self._default_embedding = None
        self._development_stage = None
        self._disease = None
        self._doi_journal = None
        self._doi_preprint = None
        self._download_url_data = None
        self._download_url_meta = None
        self._ethnicity = None
        self._feature_type = None
        self._feature_reference = None
        self.gm = None
        self._id = None
        self._individual = None
        self._organ = None
        self._organism = None
        self._primary_data = None
        self._sample_source = None
        self._sex = None
        self._source = None
        self._state_exact = None
        self._tech_sample = None
        self.treatment = None
        self._title = None
        self._year = None

        self.assay_sc_obs_key = None
        self.assay_differentiation_obs_key = None
        self.assay_type_differentiation_obs_key = None
        self.bio_sample_obs_key = None
        self.cell_line_obs_key = None
        self.cell_type_obs_key = None
        self.development_stage_obs_key = None
        self.disease_obs_key = None
        self.ethnicity_obs_key = None
        self.gm_obs_key = None
        self.individual_obs_key = None
        self.organ_obs_key = None
        self.organism_obs_key = None
        self.sample_source_obs_key = None
        self.sex_obs_key = None
        self.state_exact_obs_key = None
        self.tech_sample_obs_key = None
        self.treatment_obs_key = None

        self._feature_reference = None
        self._feature_type = None

        self.feature_id_var_key = None
        self.feature_reference_var_key = None
        self.feature_symbol_var_key = None
        self.feature_type_var_key = None

        self.spatial_x_coord_obs_key = None
        self.spatial_y_coord_obs_key = None
        self.spatial_z_coord_obs_key = None

        self.vdj_vj_1_obs_key_prefix = None
        self.vdj_vj_2_obs_key_prefix = None
        self.vdj_vdj_1_obs_key_prefix = None
        self.vdj_vdj_2_obs_key_prefix = None
        self.vdj_c_call_obs_key_suffix = None
        self.vdj_consensus_count_obs_key_suffix = None
        self.vdj_d_call_obs_key_suffix = None
        self.vdj_duplicate_count_obs_key_suffix = None
        self.vdj_j_call_obs_key_suffix = None
        self.vdj_junction_obs_key_suffix = None
        self.vdj_junction_aa_obs_key_suffix = None
        self.vdj_locus_obs_key_suffix = None
        self.vdj_productive_obs_key_suffix = None
        self.vdj_v_call_obs_key_suffix = None

        self._celltype_universe = None
        self._ontology_class_maps = dict([(k, None) for k in self._adata_ids.ontology_constrained])

        self.load_raw = None
        self.mapped_features = None
        self.remove_gene_version = None
        self.subset_gene_type = None
        self.streamlined_meta = False

        self.sample_fn = sample_fn
        self._sample_fns = sample_fns

        # Check if YAML files exists, read meta data from there if available:
        if yaml_path is not None:
            assert os.path.exists(yaml_path), f"did not find yaml {yaml_path}"
            yaml_vals = read_yaml(fn=yaml_path)
            # Set organism first as this is required to disambiguate valid entries for other meta data.
            k = "organism"
            v = yaml_vals["attr"]["organism"]
            setattr(self, k, v)
            for k, v in yaml_vals["attr"].items():
                if v is not None and k not in ["organism", "sample_fns", "dataset_index"]:
                    if isinstance(v, dict):  # v is a dictionary over file-wise meta-data items
                        assert self.sample_fn in v.keys(), f"did not find key {self.sample_fn} in yamls keys for {k}"
                        v = v[self.sample_fn]
                    # Catches spelling errors in meta data definition (yaml keys).
                    if not hasattr(self, k) and not hasattr(self, "_" + k):
                        raise ValueError(f"Tried setting unavailable property {k}.")
                    try:
                        setattr(self, k, v)
                    except AttributeError as e:
                        raise ValueError(f"An error occured when setting {k} as {v}: {e}")
            # ID can be set now already because YAML was used as input instead of child class constructor.
            self.set_dataset_id(idx=yaml_vals["meta"]["dataset_index"])

        self.load_func = load_func
        self.dict_load_func_annotation = dict_load_func_annotation
        self._additional_annotation_key = additional_annotation_key

        self.supplier = "sfaira"

    @property
    def _directory_formatted_id(self) -> str:
        return "_".join("_".join(self.id.split("/")).split("."))

    def clear(self):
        """
        Remove loaded .adata to reduce memory footprint.

        :return:
        """
        import gc
        self.adata = None
        gc.collect()

    def download(self, **kwargs):
        assert self.download_url_data is not None, f"The `download_url_data` attribute of dataset {self.id} " \
                                                   f"is not set, cannot download dataset."
        assert self.data_dir_base is not None, "No path was provided when instantiating the dataset container, " \
                                               "cannot download datasets."

        if not os.path.exists(os.path.join(self.data_dir_base, self.directory_formatted_doi)):
            os.makedirs(os.path.join(self.data_dir_base, self.directory_formatted_doi))

        urls = self.download_url_data[0] + self.download_url_meta[0]

        for url in urls:
            if url is None:
                continue
            # Special case for data that is not publically available
            if url.split(",")[0] == 'private':
                if "," in url:
                    fn = ','.join(url.split(',')[1:])
                    if os.path.isfile(os.path.join(self.data_dir, fn)):
                        print(f"File {fn} already found on disk, skipping download.")
                    else:
                        warnings.warn(f"Dataset {self.id} is not available for automatic download, please manually "
                                      f"copy the file {fn} to the following location: "
                                      f"{self.data_dir}")
                else:
                    warnings.warn(f"A file for dataset {self.id} is not available for automatic download, please"
                                  f"manually copy the associated file to the following location: {self.data_dir}")
            # Special case for data from the synapse portal
            elif url.split(",")[0].startswith('syn'):
                fn = ",".join(url.split(",")[1:])
                if os.path.isfile(os.path.join(self.data_dir, fn)):
                    print(f"File {fn} already found on disk, skipping download.")
                else:
                    self._download_synapse(url.split(",")[0], fn, **kwargs)
            # Special case for public data that is labelled as not automatically downloadable
            elif url.split(",")[0] == 'manual':
                u = ",".join(url.split(",")[2:])
                fn = url.split(",")[1]
                if os.path.isfile(os.path.join(self.data_dir, fn)):
                    print(f"File {fn} already found on disk, skipping download.")
                else:
                    print(f"Data file {fn} for dataset {self.id} cannot be retrieved automatically. "
                          f"Please download it from {u} and copy to {os.path.join(self.data_dir, fn)}")
            # All other cases
            else:
                url = urllib.parse.unquote(url)
                try:
                    urllib.request.urlopen(url)
                except urllib.error.HTTPError as err:
                    # modify headers if urllib useragent is blocked (eg.10x datasets)
                    if err.code == 403:
                        opener = urllib.request.build_opener()
                        opener.addheaders = [('User-Agent', 'Mozilla/5.0 (Windows NT 6.1; WOW64)')]
                        urllib.request.install_opener(opener)
                except urllib.error.URLError:
                    # Catch SSLCertVerificationError: [SSL: CERTIFICATE_VERIFY_FAILED] certificate verify failed: unable
                    # to get local issuer certificate (_ssl.c:1124)
                    ssl._create_default_https_context = ssl._create_unverified_context

                if 'Content-Disposition' in urllib.request.urlopen(url).info().keys():
                    fn = cgi.parse_header(urllib.request.urlopen(url).info()['Content-Disposition'])[1]["filename"]
                else:
                    fn = url.split("/")[-1]
                # Only download if file not already downloaded:
                if not os.path.isfile(os.path.join(self.data_dir, fn)):
                    print(f"Downloading: {fn}")
                    urllib.request.urlretrieve(url, os.path.join(self.data_dir, fn))

    def _download_synapse(self, synapse_entity, fn, **kwargs):
        try:
            import synapseclient
        except ImportError:
            warnings.warn("synapseclient python package not found. This package is required to download some of the "
                          "selected datasets. Run `pip install synapseclient` to install it. Skipping download of the "
                          f"following dataset: {self.id}")
            return
        import logging
        logging.captureWarnings(False)  # required to properly display warning messages below with sypaseclient loaded

        if "synapse_user" not in kwargs.keys():
            warnings.warn(f"No synapse username provided, skipping download of synapse dataset {fn}."
                          f"Provide your synapse username as the `synapse_user` argument to the download method.")
            return
        if "synapse_pw" not in kwargs.keys():
            warnings.warn(f"No synapse password provided, skipping download of synapse dataset {fn}."
                          f"Provide your synapse password as the `synapse_pw` argument to the download method.")
            return

        print(f"Downloading from synapse: {fn}")
        syn = synapseclient.Synapse()
        syn.login(kwargs['synapse_user'], kwargs['synapse_pw'])
        syn.get(entity=synapse_entity, downloadLocation=os.path.join(self.data_dir, fn))
        syn.logout()

    @property
    def cache_fn(self):
        if self.directory_formatted_doi is None or self._directory_formatted_id is None:
            # TODO is this case necessary?
            warnings.warn("Caching enabled, but Dataset.id or Dataset.doi not set. Disabling caching for now.")
            return None
        else:
            if self.cache_path is None:
                cache = self.data_dir
            else:
                cache = os.path.join(self.cache_path, self.directory_formatted_doi)
            return os.path.join(cache, "cache", self._directory_formatted_id + ".h5ad")

    def load(
            self,
            load_raw: bool = False,
            allow_caching: bool = True,
            **kwargs
    ):
        """
        Load the selected datasets into memory.
        Cache is written into director named after doi and h5ad named after data set id.
        Cache is not over-written.

        :param load_raw: Force reading the raw object even when a cached one is present.
        :param allow_caching: Write the object to cache after loading if no cache exists yet.
        """
        # Sanity checks
        if self.adata is not None:
            raise ValueError(f"adata of {self.id} already loaded.")
        if self.data_dir is None:
            raise ValueError("No sfaira data repo path provided in constructor.")

        def _error_buffered_reading(**load_kwargs):
            self.adata = self.load_func(data_dir=self.data_dir, sample_fn=self.sample_fn, **load_kwargs)

        # Run data set-specific loading script:
        def _assembly_wrapper():
            if self.load_func is None:
                raise ValueError(f"Tried to access load_func for {self.id} but did not find any.")
            _error_buffered_reading(**kwargs)
            # Enable loading of additional annotation, e.g. secondary cell type annotation
            # The additional annotation `obs2 needs to be on a subset of the original annotation `self.adata.obs`.
            if self.dict_load_func_annotation is not None:
                obs2 = self.dict_load_func_annotation[self.additional_annotation_key](
                    data_dir=self.data_dir, sample_fn=self.sample_fn)
                assert np.all([x in self.adata.obs.index for x in obs2.index]), \
                    "index mismatch between additional annotation and original"
                self.adata = self.adata[obs2.index, :]
                # Overwrite annotation
                for k, v in obs2.items():
                    self.adata.obs[k] = v

        def _cached_reading(filename):
            if filename is not None:
                if os.path.exists(filename):
                    self.adata = anndata.read_h5ad(filename)
                else:
                    _error_buffered_reading()
            else:
                _error_buffered_reading()

        def _cached_writing(filename):
            if filename is not None:
                dir_cache = os.path.dirname(filename)
                if not os.path.exists(dir_cache):
                    os.makedirs(dir_cache)
                if not os.path.exists(filename) and self.adata is not None:
                    self.adata.write_h5ad(filename)

        if load_raw and allow_caching:
            _assembly_wrapper()
            _cached_writing(self.cache_fn)
        elif load_raw and not allow_caching:
            _assembly_wrapper()
        elif not load_raw and allow_caching:
            _cached_reading(self.cache_fn)
            _cached_writing(self.cache_fn)
        else:  # not load_raw and not allow_caching
            _cached_reading(self.cache_fn)

        # Set loading-specific metadata:
        self.load_raw = load_raw

    load.__doc__ = load_doc

    def _add_missing_featurenames(
            self,
            match_to_reference: Union[str, bool, None],
    ):
        if self.feature_symbol_var_key is None and self.feature_id_var_key is None:
            raise ValueError("Either feature_symbol_var_key or feature_id_var_key needs to be provided in the"
                             " dataloader")
        elif self.feature_symbol_var_key is None and self.feature_id_var_key:
            # Convert ensembl ids to gene symbols
            id_dict = self.genome_container.id_to_symbols_dict
            ensids = self.adata.var.index if self.feature_id_var_key == "index" else \
                self.adata.var[self.feature_id_var_key]
            self.adata.var[self._adata_ids.feature_symbol] = [
                id_dict[n.split(".")[0]] if n.split(".")[0] in id_dict.keys() else 'n/a'
                for n in ensids
            ]
            if self.adata.raw is not None:
                ensids = self.adata.raw.var.index if self.feature_id_var_key == "index" else \
                    self.adata.raw.var[self.feature_id_var_key]
                self.adata.raw.var[self._adata_ids.feature_symbol] = [
                    id_dict[n.split(".")[0]] if n.split(".")[0] in id_dict.keys() else 'n/a'
                    for n in ensids
                ]
            self.feature_symbol_var_key = self._adata_ids.feature_symbol
        elif self.feature_symbol_var_key and self.feature_id_var_key is None:
            # Convert gene symbols to ensembl ids
            id_dict = self.genome_container.symbol_to_id_dict
            id_strip_dict = self.genome_container.strippednames_to_id_dict
            # Matching gene names to ensembl ids in the following way: if the gene is present in the ensembl dictionary,
            # match it straight away, if it is not in there we try to match everything in front of the first period in
            # the gene name with a dictionary that was modified in the same way, if there is still no match we append na
            ensids = []
            symbs = self.adata.var.index if self.feature_symbol_var_key == "index" else \
                self.adata.var[self.feature_symbol_var_key]
            for n in symbs:
                if n in id_dict.keys():
                    ensids.append(id_dict[n])
                elif n.split(".")[0] in id_strip_dict.keys():
                    ensids.append(id_strip_dict[n.split(".")[0]])
                else:
                    ensids.append('n/a')
            self.adata.var[self._adata_ids.feature_id] = ensids
            if self.adata.raw is not None:
                ensids = []
                symbs = self.adata.raw.var.index if self.feature_symbol_var_key == "index" else \
                    self.adata.raw.var[self.feature_symbol_var_key]
                for n in symbs:
                    if n in id_dict.keys():
                        ensids.append(id_dict[n])
                    elif n.split(".")[0] in id_strip_dict.keys():
                        ensids.append(id_strip_dict[n.split(".")[0]])
                    else:
                        ensids.append('n/a')
                self.adata.raw.var[self._adata_ids.feature_id] = ensids
            self.feature_id_var_key = self._adata_ids.feature_id

    def _collapse_ensembl_gene_id_versions(self):
        """
        Remove version tag on ensembl gene ID so that different versions are superimposed downstream.

        :return:
        """
        if not self.feature_id_var_key:
            raise ValueError(
                "Cannot remove gene version when gene_id_ensembl_var_key is not set in dataloader and "
                "match_to_reference is False"
            )
        elif self.feature_id_var_key == "index":
            self.adata.index = [x.split(".")[0] for x in self.adata.var.index]
            if self.adata.raw is not None:
                self.adata.raw.index = [x.split(".")[0] for x in self.adata.raw.var.index]
        else:
            self.adata.var[self.feature_id_var_key] = [
                x.split(".")[0] for x in self.adata.var[self.feature_id_var_key].values
            ]
            if self.adata.raw is not None:
                self.adata.raw.var[self.feature_id_var_key] = [
                    x.split(".")[0] for x in self.adata.raw.var[self.feature_id_var_key].values
                ]
        # Collapse if necessary:
        self.adata = collapse_matrix(adata=self.adata, var_column=self.feature_id_var_key)
        if self.adata.raw is not None:
            self.adata.raw = collapse_matrix(adata=self.adata.raw, var_column=self.feature_id_var_key)

    def collapse_counts(self):
        """
        Collapse count matrix along duplicated index.
        """
        if len(np.unique(self.adata.var.index)) < self.adata.var.shape[0]:
            self.adata = collapse_matrix(adata=self.adata, var_column="index")

    def streamline_features(
            self,
            match_to_release: Union[str, Dict[str, str], None] = None,
            remove_gene_version: bool = True,
            subset_genes_to_type: Union[None, str, List[str]] = None,
            schema: Union[str, None] = None,
    ):
        """
        Subset and sort genes to genes defined in an assembly or genes of a particular type, such as protein coding.
        This also adds missing ensid or gene symbol columns if match_to_reference is not set to False and removes all
        adata.var columns that are not defined as gene_id_ensembl_var_key or gene_id_symbol_var_key in the dataloader.

        :param match_to_release: Which genome annotation release to map the feature space to. Note that assemblies from
            ensbeml are usually named as Organism.Assembly.Release, this is the Release string. Can be:
                - str: Provide the name of the release.
                - dict: Mapping of organism to name of the release (see str format). Chooses release for each
                    data set based on organism annotation.
        :param remove_gene_version: Whether to remove the version number after the colon sometimes found in ensembl
            gene ids.
        :param subset_genes_to_type: Type(s) to subset to. Can be a single type or a list of types or None.
            Types can be:
                - None: All genes in assembly.
                - "protein_coding": All protein coding genes in assembly.
        """
        self.__assert_loaded()
        if schema is not None:
            # schema_version = schema.split(":")[-1] if ":" in schema else None
            # Set schema as provided by the user
            if schema.startswith("sfaira"):
                adata_target_ids = AdataIdsSfaira()
            elif schema.startswith("cellxgene"):
                adata_target_ids = AdataIdsCellxgene_v2_0_0()
            else:
                raise ValueError(f"did not recognize schema {schema}")
            match_to_release = adata_target_ids.feature_kwargs["match_to_release"]
            remove_gene_version = adata_target_ids.feature_kwargs["remove_gene_version"]
            subset_genes_to_type = adata_target_ids.feature_kwargs["subset_genes_to_type"]

        # Move data matrices around.
        assert self.layer_processed or self.layer_counts
        if self.layer_processed is None:
            x_proc = None
            var_proc = None
        elif self.layer_processed == "X":
            x_proc = self.adata.X
            var_proc = self.adata.var
        elif self.layer_processed == "raw":
            x_proc = self.adata.raw.X
            var_proc = self.adata.raw.var
        else:
            if self.layer_processed not in self.adata.layers.keys():
                raise ValueError(f"layer {self.layer_processed} not in layers {list(self.adata.layers.keys())}")
            else:
                x_proc = self.adata.layers[self.layer_processed]
                var_proc = self.adata.var
        if self.layer_counts is None:
            x_counts = None
            var_counts = None
        elif self.layer_counts == "X":
            x_counts = self.adata.X
            var_counts = self.adata.var
        elif self.layer_counts == "raw":
            x_counts = self.adata.raw.X
            var_counts = self.adata.raw.var
        else:
            if self.layer_counts not in self.adata.layers.keys():
                raise ValueError(f"layer {self.layer_counts} not in layers {list(self.adata.layers.keys())}")
            else:
                x_counts = self.adata.layers[self.layer_counts]
                var_counts = self.adata.var
        if x_proc is None and x_counts is not None:
            x = x_counts
            var = var_counts
            x_raw = None
            var_raw = None
        elif x_proc is not None and x_counts is None:
            x = x_proc
            var = var_proc
            x_raw = None
            var_raw = None
        elif x_proc is not None and x_counts is not None:
            x = x_proc
            var = var_proc
            x_raw = x_counts
            var_raw = var_counts
        else:
            assert False
        self.adata = anndata.AnnData(
            X=x,
            obs=self.adata.obs,
            obsm=self.adata.obsm,
            var=var,
            uns=self.adata.uns
        )
        if x_raw is not None:
            self.adata.raw = self.adata = anndata.AnnData(
                X=x_raw,
                var=var_raw
            )

        # Set genome container if mapping of gene labels is requested
        if isinstance(match_to_release, dict):
            match_to_release = match_to_release[self.organism]
        self._set_genome(organism=self.organism, release=match_to_release)
        self.mapped_features = self.genome_container.release

        self.remove_gene_version = remove_gene_version
        self.subset_gene_type = subset_genes_to_type
        # Streamline feature space:
        self._add_missing_featurenames(match_to_reference=match_to_release)
        for key in [self.feature_id_var_key, self.feature_symbol_var_key]:
            # Make features unique (to avoid na-matches in converted columns to be collapsed by
            # _collapse_ensembl_gene_id_versions() below.
            if not key:
                pass
            elif key == "index":
                self.adata.var.index = make_index_unique(self.adata.var.index).tolist()
            else:
                self.adata.var[key] = make_index_unique(pd.Index(self.adata.var[key].values.tolist())).tolist()
        if remove_gene_version:
            self._collapse_ensembl_gene_id_versions()

        # Compute indices of genes to keep
        if subset_genes_to_type is None:
            subset_ids_ensg = self.genome_container.ensembl
            subset_ids_symbol = self.genome_container.symbols
        else:
            if isinstance(subset_genes_to_type, str):
                subset_genes_to_type = [subset_genes_to_type]
            keys = np.unique(self.genome_container.biotype)
            if subset_genes_to_type not in keys:
                raise ValueError(f"subset type {subset_genes_to_type} not available in list {keys}")
            subset_ids_ensg = [
                x.upper() for x, y in zip(self.genome_container.ensembl, self.genome_container.biotype)
                if y in subset_genes_to_type
            ]
            subset_ids_symbol = [
                x for x, y in zip(self.genome_container.symbols, self.genome_container.biotype)
                if y in subset_genes_to_type
            ]

        x_new, var_new = subset_adata_genes(
            x=self.adata.X,
            var=self.adata.var,
            feature_id_var_key=self.feature_id_var_key,
            feature_symbol_var_key=self.feature_symbol_var_key,
            subset_ids_ensg=subset_ids_ensg,
            subset_ids_symbol=subset_ids_symbol)
        adata_new = anndata.AnnData(
            X=x_new,
            obs=self.adata.obs,
            obsm=self.adata.obsm,
            var=var_new,
            uns=self.adata.uns
        )
        adata_new.uns[self._adata_ids.mapped_features] = match_to_release
        if self.adata.raw is not None:
            x_raw_new, var_raw_new = subset_adata_genes(
                x=self.adata.raw.X,
                var=self.adata.raw.var,
                feature_id_var_key=self.feature_id_var_key,
                feature_symbol_var_key=self.feature_symbol_var_key,
                subset_ids_ensg=subset_ids_ensg,
                subset_ids_symbol=subset_ids_symbol)
            adata_new.raw = anndata.AnnData(
                X=x_raw_new,
                obs=self.adata.obs,
                var=var_raw_new,
            )
        self.adata = adata_new

    def streamline_metadata(
            self,
            schema: str = "sfaira",
            clean_obs: bool = True,
            clean_var: bool = True,
            clean_uns: bool = True,
            clean_obs_names: bool = True,
            keep_orginal_obs: bool = False,
            keep_symbol_obs: bool = True,
            keep_id_obs: bool = True,
    ):
        """
        Streamline the adata instance to a defined output schema.

        Output format are saved in ADATA_FIELDS* classes.

        Note on ontology-controlled meta data:
        These are defined for a given format in `ADATA_FIELDS*.ontology_constrained`.
        They may appear in three different formats:
            - original (free text) annotation
            - ontology symbol
            - ontology ID
        During streamlining, these ontology-controlled meta data are projected to all of these three different formats.
        The initially annotated column may be any of these and is defined as "{attr}_obs_col".
        The resulting three column per meta data item are named:
            - ontology symbol: "{ADATA_FIELDS*.attr}"
            - ontology ID: {ADATA_FIELDS*.attr}_{ADATA_FIELDS*.onto_id_suffix}"
            - original (free text) annotation: "{ADATA_FIELDS*.attr}_{ADATA_FIELDS*.onto_original_suffix}"

        :param schema: Export format.
            - "sfaira"
            - "cellxgene"
        :param clean_obs: Whether to delete non-streamlined fields in .obs, .obsm and .obsp.
        :param clean_var: Whether to delete non-streamlined fields in .var, .varm and .varp.
        :param clean_uns: Whether to delete non-streamlined fields in .uns.
        :param clean_obs_names: Whether to replace obs_names with a string comprised of dataset id and an increasing
            integer.
        :param keep_orginal_obs: For ontology-constrained .obs columns, whether to keep a column with original
            annotation.
        :param keep_symbol_obs: For ontology-constrained .obs columns, whether to keep a column with ontology symbol
            annotation.
        :param keep_id_obs: For ontology-constrained .obs columns, whether to keep a column with ontology ID annotation.
        :return:
        """
        schema_version = schema.split(":")[-1] if ":" in schema else None
        self.__assert_loaded()

        # Set schema as provided by the user
        if schema.startswith("sfaira"):
            adata_target_ids = AdataIdsSfaira()
        elif schema.startswith("cellxgene"):
            adata_target_ids = AdataIdsCellxgene_v2_0_0()
        else:
            raise ValueError(f"did not recognize schema {schema}")

        if hasattr(adata_target_ids, "gene_id_ensembl") and not hasattr(self._adata_ids, "gene_id_ensembl"):
            raise ValueError(f"Cannot convert this object to schema {schema}, as the currently applied schema does not "
                             f"have an ensembl gene ID annotation. Please run .streamline_features() first.")
        experiment_batch_labels = [getattr(self._adata_ids, x) for x in self._adata_ids.batch_keys]

        # Creating new var annotation
        var_new = pd.DataFrame()
        for k in adata_target_ids.var_keys:
            if k == "feature_id":
                if not self.feature_id_var_key:
                    raise ValueError("feature_id not set in dataloader despite being required by the "
                                     "selected meta data schema. please run streamline_features() first to create the "
                                     "missing annotation")
                elif self.feature_id_var_key == "index":
                    var_new[getattr(adata_target_ids, k)] = self.adata.var.index.tolist()
                else:
                    var_new[getattr(adata_target_ids, k)] = self.adata.var[self.feature_id_var_key].tolist()
                    del self.adata.var[self.feature_id_var_key]
                self.feature_id_var_key = getattr(adata_target_ids, k)
            elif k == "feature_symbol":
                if not self.feature_symbol_var_key:
                    raise ValueError("gene_id_symbols_var_key not set in dataloader despite being required by the "
                                     "selected meta data schema. please run streamline_features() first to create the "
                                     "missing annotation")
                elif self.feature_symbol_var_key == "index":
                    var_new[getattr(adata_target_ids, k)] = self.adata.var.index.tolist()
                else:
                    var_new[getattr(adata_target_ids, k)] = self.adata.var[self.feature_symbol_var_key].tolist()
                    del self.adata.var[self.feature_symbol_var_key]
                self.feature_symbol_var_key = getattr(adata_target_ids, k)
            else:
                val = getattr(self, k)
                while hasattr(val, '__len__') and not isinstance(val, str) and len(val) == 1:  # unpack nested lists/tuples
                    val = val[0]
                var_new[getattr(adata_target_ids, k)] = val
        # set var index
        var_new.index = var_new[adata_target_ids.feature_index].tolist()
        if clean_var:
            if self.adata.varm is not None:
                del self.adata.varm
            if self.adata.varp is not None:
                del self.adata.varp
            self.adata.var = var_new
            if "feature_id" not in adata_target_ids.var_keys:
                self.feature_id_var_key = None
            if "feature_symbol" not in adata_target_ids.var_keys:
                self.feature_symbol_var_key = None
        else:
            index_old = self.adata.var.index.copy()
            # Add old columns in if they are not duplicated:
            self.adata.var = pd.concat([
                var_new,
                pd.DataFrame(dict([(k, v) for k, v in self.adata.var.items() if k not in var_new.columns]))
            ], axis=1)
            self.adata.var.index = index_old

        # Prepare new .uns dict:
        uns_new = {}
        for k in adata_target_ids.uns_keys:
            if hasattr(self, k) and getattr(self, k) is not None:
                val = getattr(self, k)
            elif hasattr(self, f"{k}_obs_key") and getattr(self, f"{k}_obs_key") is not None:
                val = self.adata.obs[getattr(self, f"{k}_obs_key")].unique().astype(str).tolist()
            elif getattr(self._adata_ids, k) in self.adata.obs.columns:
                val = self.adata.obs[getattr(self._adata_ids, k)].unique().astype(str).tolist()
            else:
                val = None
            while hasattr(val, '__len__') and not isinstance(val, str) and len(val) == 1:  # Unpack nested lists/tuples.
                val = val[0]
            uns_new[getattr(adata_target_ids, k)] = val
        if clean_uns:
            self.adata.uns = uns_new
        else:
            self.adata.uns.update(uns_new)

        # Prepare new .obs dataframe
        # Queried meta data may be:
        # 1) in .obs
        #   a) for an ontology-constrained meta data item
        #       I) as free annotation with a term map to an ontology
        #       II) as column with ontology symbols
        #       III) as column with ontology IDs
        #   b) for a non-ontology-constrained meta data item:
        #       I) as free annotation
        # 2) in .uns
        #   b) as elements that are ontology symbols
        #   c) as elements that are ontology IDs
        # .obs annotation takes priority over .uns annotation if both are present.
        # The output columns are:
        # - for an ontology-constrained meta data item "attr":
        #   *  symbols:         "attr"
        #   *  IDs:             "attr" + self._adata_ids.onto_id_suffix
        #   *  original labels: "attr" + self._adata_ids.onto_original_suffix
        # - for a non-ontology-constrained meta data item "attr":
        #   *  original labels: "attr" + self._adata_ids.onto_original_suffix
        obs_new = pd.DataFrame(index=self.adata.obs.index)
        for k in [x for x in adata_target_ids.obs_keys]:
            if k in experiment_batch_labels and getattr(self, f"{k}_obs_key") is not None:
                # Handle batch-annotation columns which can be provided as a combination of columns separated by an
                # asterisk.
                # The queried meta data are always:
                # 1b-I) a combination of existing columns in .obs
                old_cols = getattr(self, f"{k}_obs_key")
                batch_cols = []
                for batch_col in old_cols.split("*"):
                    if batch_col in self.adata.obs_keys():
                        batch_cols.append(batch_col)
                    else:
                        # This should not occur in single data set loaders (see warning below) but can occur in
                        # streamlined data loaders if not all instances of the streamlined data sets have all columns
                        # in .obs set.
                        print(f"WARNING: attribute {batch_col} of data set {self.id} was not found in columns.")
                # Build a combination label out of all columns used to describe this group.
                # Add data set label into this label so that these groups are unique across data sets.
                val = [
                    self.doi_cleaned_id + "_".join([str(xxx) for xxx in xx])
                    for xx in zip(*[self.adata.obs[batch_col].values.tolist() for batch_col in batch_cols])
                ]
            else:
                # Locate annotation.
                if hasattr(self, f"{k}_obs_key") and getattr(self, f"{k}_obs_key") is not None and \
                        getattr(self, f"{k}_obs_key") in self.adata.obs.columns:
                    # Last and-clause to check if this column is included in data sets. This may be violated if data
                    # is obtained from a database which is not fully streamlined.
                    # Look for 1a-* and 1b-I
                    val = self.adata.obs[getattr(self, f"{k}_obs_key")].values.tolist()
                else:
                    # Look for 2a, 2b
                    val = getattr(self, k)
                    if val is None:
                        val = self._adata_ids.unknown_metadata_identifier
                    # Unpack nested lists/tuples:
                    while hasattr(val, '__len__') and not isinstance(val, str) and len(val) == 1:
                        val = val[0]
                    val = [val] * self.adata.n_obs
            # Identify annotation: disambiguate 1a-I, 1a-II, 1a-III, 1b-I.
            if k in self._adata_ids.ontology_constrained:
                # 1a-*.
                if isinstance(self.get_ontology(k=k), OntologyHierarchical) and np.all([
                    self.get_ontology(k=k).is_a_node_name(x) or x == self._adata_ids.unknown_metadata_identifier
                    for x in np.unique(val)
                ]):  # 1a-II)
                    new_col = getattr(adata_target_ids, k)
                    validation_ontology = self.get_ontology(k=k)
                elif isinstance(self.get_ontology(k=k), OntologyHierarchical) and np.all([
                    self.get_ontology(k=k).is_a_node_id(x) or x == self._adata_ids.unknown_metadata_identifier
                    for x in np.unique(val)
                ]):  # 1a-III)
                    new_col = getattr(adata_target_ids, k) + self._adata_ids.onto_id_suffix
                    validation_ontology = None
                else:  # 1a-I)
                    new_col = getattr(adata_target_ids, k) + self._adata_ids.onto_original_suffix
                    validation_ontology = None
            else:
                # 1b-I.
                new_col = getattr(adata_target_ids, k)
                validation_ontology = self.get_ontology(k=k)
            # Check values for validity:
            self._value_protection(attr=new_col, allowed=validation_ontology, attempted=[
                x for x in np.unique(val)
                if x not in [
                    self._adata_ids.unknown_metadata_identifier,
                ]
            ])
            obs_new[new_col] = val
            # For ontology-constrained meta data, the remaining columns are added after .obs cleaning below.
        if clean_obs:
            if self.adata.obsm is not None:
                del self.adata.obsm
            if self.adata.obsp is not None:
                del self.adata.obsp
            self.adata.obs = obs_new
        else:
            index_old = self.adata.obs.index.copy()
            # Add old columns in if they are not duplicated in target obs column space, even if this column is not
            # defined. This would result in the instance accessing this column assuming it was streamlined.
            self.adata.obs = pd.concat([
                obs_new,
                pd.DataFrame(dict([(k, v) for k, v in self.adata.obs.items()
                                   if k not in adata_target_ids.controlled_meta_keys]))
            ], axis=1)
            self.adata.obs.index = index_old
        for k in [x for x in adata_target_ids.obs_keys if x in adata_target_ids.ontology_constrained]:
            # Add remaining output columns for ontology-constrained meta data.
            self.__impute_ontology_cols_obs(attr=k, adata_ids=adata_target_ids)
            # Delete attribute-specific columns that are not desired.
            col_name = getattr(self._adata_ids, k) + self._adata_ids.onto_id_suffix
            if not keep_id_obs and col_name in self.adata.obs.columns:
                del self.adata.obs[col_name]
            col_name = getattr(self._adata_ids, k) + self._adata_ids.onto_original_suffix
            if not keep_orginal_obs and col_name in self.adata.obs.columns:
                del self.adata.obs[col_name]
            col_name = getattr(self._adata_ids, k)
            if not keep_symbol_obs and col_name in self.adata.obs.columns:
                del self.adata.obs[col_name]
        if clean_obs_names:
            self.adata.obs.index = [f"{self.id}_{i}" for i in range(1, self.adata.n_obs + 1)]

        # Make sure that correct unknown_metadata_identifier is used in .uns, .obs and .var metadata
        unknown_old = self._adata_ids.unknown_metadata_identifier
        unknown_new = adata_target_ids.unknown_metadata_identifier
        self.adata.obs = self.adata.obs.replace({None: unknown_new})
        self.adata.obs = self.adata.obs.replace({unknown_old: unknown_new})
        self.adata.var = self.adata.var.replace({None: unknown_new})
        self.adata.var = self.adata.var.replace({unknown_old: unknown_new})
        for k in self.adata.uns_keys():
            if self.adata.uns[k] is None or self.adata.uns[k] == unknown_old:
                self.adata.uns[k] = unknown_new

        self._adata_ids = adata_target_ids  # set new adata fields to class after conversion
        self.streamlined_meta = True
        # Add additional hard-coded description changes for cellxgene schema:
        if schema.startswith("cellxgene"):
            self.adata = cellxgene_export_adaptor(adata=self.adata, adata_ids=self._adata_ids, version=schema_version,
                                                  obs_keys_batch=self.tech_sample_obs_key)

    def write_distributed_store(
            self,
            dir_cache: Union[str, os.PathLike],
            store_format: str = "dao",
            dense: bool = False,
            compression_kwargs: Union[dict, None] = None,
            chunks: Union[int, None] = None,
            shuffle_data: bool = False
    ):
        """
        Write data set into a format that allows distributed access to data set on disk.

        Stores are useful for distributed access to data sets, in many settings this requires some streamlining of the
        data sets that are accessed. Use .streamline_* before calling this method to streamline the data sets.

        :param dir_cache: Directory to write cache in.
        :param store_format: Disk format for objects in cache. Recommended is "dao".

            - "h5ad": Allows access via backed .h5ad.
                Note on compression: .h5ad supports sparse data with is a good compression that gives fast row-wise
                    access if the files are csr, so further compression potentially not necessary.
            - "dao": Distributed access optimised format, recommended for batched access in optimisation, for example.
        :param dense: Whether to write sparse or dense store, this will be homogenously enforced.
        :param compression_kwargs: Compression key word arguments to give to h5py or zarr
            For store_format=="h5ad", see also anndata.AnnData.write_h5ad:
                - compression,
                - compression_opts.
            For store_format=="dao", see also sfaira.data.write_dao which relays kwargs to
            zarr.hierarchy.create_dataset:
                - compressor
                - overwrite
                - order
                and others.
        :param chunks: Observation axes of chunk size of zarr array, see anndata.AnnData.write_zarr documentation.
            Only relevant for store=="dao". The feature dimension of the chunks is always is the full feature space.
            Uses zarr default chunking across both axes if None.
        :param shuffle_data: If True -> shuffle ordering of cells in datasets before writing store
        """
        if compression_kwargs is None:
            compression_kwargs = {}
        self.__assert_loaded()
        if store_format == "h5ad":
            if not isinstance(self.adata.X, scipy.sparse.csr_matrix):
                print(f"WARNING: high-perfomances caches based on .h5ad work better with .csr formatted expression "
                      f"data, found {type(self.adata.X)}")
            fn = os.path.join(dir_cache, self.doi_cleaned_id + ".h5ad")
            as_dense = ("X",) if dense else ()
            print(f"writing {self.adata.shape} into {fn}")
            self.adata.write_h5ad(filename=fn, as_dense=as_dense, **compression_kwargs)
        elif store_format == "dao":
            # Convert data object to sparse / dense as required:
            if not dense:
                raise ValueError("WARNING: sparse zarr array performance is not be optimal and not supported yet, "
                                 "consider writing as dense and consider that zarr arrays are compressed on disk!")
            fn = os.path.join(dir_cache, self.doi_cleaned_id)
            chunks = (chunks, self.adata.X.shape[1]) if chunks is not None else True
            write_dao(store=fn, adata=self.adata, chunks=chunks, compression_kwargs=compression_kwargs,
                      shuffle_data=shuffle_data)
        else:
            raise ValueError()

    def _set_genome(self, organism: str, release: str):
        self.genome_container = GenomeContainer(organism=organism, release=release)

    @property
    def doi_cleaned_id(self):
        return "_".join(self.id.split("_")[:-1])

    def get_ontology(self, k) -> Union[OntologyHierarchical, None]:
        x = getattr(self.ontology_container_sfaira, k) if hasattr(self.ontology_container_sfaira, k) else None
        if x is not None and isinstance(x, dict):
            assert isinstance(self.organism, str), self.organism
            # Check if organism-specific option is available, otherwise choose generic option:
            if self.organism in x.keys():
                k = self.organism
            else:
                k = self.ontology_container_sfaira.key_other
                assert k in x.keys(), x.keys()  # Sanity check on dictionary keys.
            x = x[k]
            assert x is None or isinstance(x, Ontology), x  # Sanity check on dictionary element.
        return x

    def _write_ontology_class_map(self, fn, tab: pd.DataFrame):
        """
        Write class map to file.

        Helper to allow direct interaction with written table instead of using table from instance.

        :param fn: File name of csv to write class maps to.
        :param tab: Class map table.
        """
        tab.to_csv(fn, index=False, sep="\t")

    def write_ontology_class_maps(
            self,
            fn,
            attrs: List[str],
            protected_writing: bool = True,
            **kwargs
    ):
        """
        Load class maps of ontology-controlled field to ontology classes.

        TODO: deprecate and only keep DatasetGroup writing?

        :param fn: File name of tsv to write class maps to.
        :param attrs: Attributes to create a tsv for. Must correspond to *_obs_key in yaml.
        :param protected_writing: Only write if file was not already found.
        :return:
        """
        for x in attrs:
            k = getattr(self, x + "_obs_key")
            if k is None:
                warnings.warn(f"attempted to write ontology class maps for data set {self.id} without annotation of "
                              f"meta data {x}")
            elif k not in self.adata.obs.columns:
                warnings.warn(f"attempted to write ontology class maps for data set {self.id} but did not find column "
                              f"{k} in .obs which should correspond to {x}")
            else:
                fn_x = fn + "_" + x + ".tsv"
                labels_original = np.unique(self.adata.obs[k].values)
                tab = prepare_ontology_map_tab(
                    source=labels_original,
                    onto=x,
                    organism=self.organism,
                    match_only=False,
                    anatomical_constraint=self.organ,
                    include_synonyms=True,
                    omit_list=[self._adata_ids.unknown_metadata_identifier],
                    **kwargs
                )
                if not os.path.exists(fn_x) or not protected_writing:
                    self._write_ontology_class_map(fn=fn_x, tab=tab)

    def _read_ontology_class_map(self, fn) -> pd.DataFrame:
        """
        Read class map.

        Helper to allow direct interaction with resulting table instead of loading into instance.

        :param fn: File name of csv to load class maps from.
        :return:
        """
        try:
            # Need dtype="str" to force numeric cell type identifiers, e.g. cluster numbers to be in string format.
            tab = pd.read_csv(fn, header=0, index_col=None, sep="\t", dtype="str")
        except pandas.errors.ParserError as e:
            print(f"{self.id}")
            raise pandas.errors.ParserError(e)
        return tab

    def read_ontology_class_maps(self, fns: List[str]):
        """
        Load class maps of free text class labels to ontology classes.

        :param fns: File names of tsv to load class maps from.
        :return:
        """
        assert isinstance(fns, list)
        ontology_class_maps = {}
        for fn in fns:
            k = identify_tsv(fn=fn, ontology_names=self._adata_ids.ontology_constrained)
            if os.path.exists(fn):
                ontology_class_maps[k] = self._read_ontology_class_map(fn=fn)
        self.ontology_class_maps = ontology_class_maps

    def project_free_to_ontology(self, attr: str):
        """
        Project free text cell type names to ontology based on mapping table.

        ToDo: add ontology ID setting here.
        """
        ontology_map = self.ontology_class_maps[attr]
        adata_fields = self._adata_ids
        col_original = attr + adata_fields.onto_original_suffix
        labels_original = self.adata.obs[col_original].values
        if ontology_map is not None:  # only if this was defined
            labels_mapped = [
                ontology_map[x] if x in ontology_map.keys()
                else x for x in labels_original
            ]
            # Convert unknown celltype placeholders (needs to be hardcoded here as placeholders are also hardcoded in
            # conversion tsv files
            placeholder_conversion = {
                "UNKNOWN": adata_fields.unknown_metadata_identifier,
                "NOT_A_CELL": adata_fields.not_a_cell_celltype_identifier,
            }
            labels_mapped = [
                placeholder_conversion[x] if x in placeholder_conversion.keys()
                else x for x in labels_mapped
            ]
            map_exceptions = [adata_fields.unknown_metadata_identifier]
            if attr == "cell_type":
                map_exceptions.append(adata_fields.not_a_cell_celltype_identifier)
            # Validate mapped IDs based on ontology:
            # This aborts with a readable error if there was a target in the mapping file that doesnt match the ontology
            # This protection blocks progression in the unit test if not deactivated.
            self._value_protection(
                attr=attr,
                allowed=self.get_ontology(k=attr),
                attempted=[x for x in list(set(labels_mapped)) if x not in map_exceptions],
            )
            # Add cell type IDs into object:
            # The IDs are not read from a source file but inferred based on the class name.
            # TODO this could be changed in the future, this allows this function to be used both on cell type name
            #  mapping files with and without the ID in the third column.
            # This mapping blocks progression in the unit test if not deactivated.
            self.adata.obs[getattr(adata_fields, attr)] = labels_mapped
            self.__project_ontology_ids_obs(attr=attr, map_exceptions=map_exceptions, from_id=False,
                                            adata_ids=adata_fields)
        else:
            # Assumes that the original labels are the correct ontology symbols, because of a lack of ontology,
            # ontology IDs cannot be inferred.
            # TODO is this necessary in the future?
            self.adata.obs[getattr(adata_fields, attr)] = labels_original
            self.adata.obs[getattr(adata_fields, attr) + adata_fields.onto_id_suffix] = \
                [adata_fields.unknown_metadata_identifier] * self.adata.n_obs
        self.adata.obs[getattr(adata_fields, attr) + adata_fields.onto_original_suffix] = labels_original

    def __impute_ontology_cols_obs(
            self,
            attr: str,
            adata_ids: AdataIds,
    ):
        """
        Add missing ontology defined columns (symbol, ID, original) for a given ontology.

        1) If original column is non-empty and symbol and ID are empty:
            orginal column is projected to ontology and both symbol and ID are inferred.
            Note that in this case, a label map is required.
        2) If ID column is non-empty or symbol is non-empty, an error is thrown.
            a) If ID column is non-empty and symbol is empty, symbol is inferred.
            b) If ID column is empty and symbol is non-empty, ID is inferred.
            c) If ID column is non-empty and non-symbol is empty, symbol is inferred and over-written.
                Note that this setting allows usage of data sets which were streamlined with a different ontology
                version.
            In all cases original is kept if it is set and is set to symbol otherwise.
        3) If original, ID and symbol columns are empty, no action is taken (meta data item was not set).
        """
        ontology = self.get_ontology(k=attr)
        col_symbol = getattr(adata_ids, attr)
        col_id = getattr(adata_ids, attr) + self._adata_ids.onto_id_suffix
        col_original = getattr(adata_ids, attr) + self._adata_ids.onto_original_suffix
        if ontology is None:
            # Fill with invalid ontology identifiers if no ontology was found.
            self.adata.obs[col_id] = \
                [self._adata_ids.invalid_metadata_identifier for _ in range(self.adata.n_obs)]
            self.adata.obs[col_original] = \
                [self._adata_ids.invalid_metadata_identifier for _ in range(self.adata.n_obs)]
            self.adata.obs[col_symbol] = \
                [self._adata_ids.invalid_metadata_identifier for _ in range(self.adata.n_obs)]
        else:
            # Note that for symbol and ID, the columns may be filled but not streamlined according to the ontology,
            # in that case the corresponding meta data is defined as absent.
            # Check which level of meta data annotation is present.
            # Symbols:
            symbol_col_present = col_symbol in self.adata.obs.columns
            symbol_col_streamlined = np.all([
                ontology.is_a_node_name(x) or x == self._adata_ids.unknown_metadata_identifier
                for x in np.unique(self.adata.obs[col_symbol].values)]) if symbol_col_present else False
            symbol_present = symbol_col_present and symbol_col_streamlined
            # IDs:
            id_col_present = col_id in self.adata.obs.columns
            id_col_streamlined = np.all([
                ontology.is_a_node_id(x) or x == self._adata_ids.unknown_metadata_identifier
                for x in np.unique(self.adata.obs[col_id].values)]) if id_col_present else False
            id_present = id_col_present and id_col_streamlined
            # Original annotation (free text):
            original_present = col_original in self.adata.obs.columns
            if original_present and not symbol_present and not id_present:  # 1)
                self.project_free_to_ontology(attr=attr)
            if symbol_present or id_present:  # 2)
                if symbol_present and not id_present:  # 2a)
                    self.__project_ontology_ids_obs(attr=attr, from_id=False, adata_ids=adata_ids)
                if not symbol_present and id_present:  # 2b)
                    self.__project_ontology_ids_obs(attr=attr, from_id=True, adata_ids=adata_ids)
                if symbol_present and id_present:  # 2c)
                    self.__project_ontology_ids_obs(attr=attr, from_id=True, adata_ids=adata_ids)
                if not original_present:
                    val = self.adata.obs[col_symbol]
                    self.adata.obs[col_original] = val

    def __project_ontology_ids_obs(
            self,
            attr: str,
            adata_ids: AdataIds,
            map_exceptions: Union[None, List[str]] = None,
            map_exceptions_value=None,
            from_id: bool = False,
    ):
        """
        Project ontology names to IDs for a given ontology in .obs entries.

        :param attr: name of obs_column containing names to convert or python list containing these values
        :param map_exceptions: list of values that should not be mapped.
            Defaults to unknown meta data identifier defined in ID object if None.
        :param map_exceptions_value: placeholder target value for values excluded from mapping.
            Defaults to unknown meta data identifier defined in ID object if None.
        :param from_id: Whether to output ontology symbol or ID.
        :return:
        """
        ontology = self.get_ontology(k=attr)
        assert ontology is not None, f"cannot project value for {attr} because ontology is None"
        assert isinstance(attr, (str, list)), f"argument key_in needs to be of type str or list. Supplied" \
                                              f"type: {type(attr)}"
        map_exceptions = map_exceptions if map_exceptions is not None else [adata_ids.unknown_metadata_identifier]
        map_exceptions = [x.lower() for x in map_exceptions]
        if map_exceptions_value is None:
            # TODO this may be simplified in the future once all unknown meta data labels are the same.
            if attr == "cell_type":
                map_exceptions_value = adata_ids.unknown_metadata_identifier
            else:
                map_exceptions_value = adata_ids.unknown_metadata_identifier
        col_name = getattr(adata_ids, attr)
        if from_id:
            col_name += adata_ids.onto_id_suffix
        input_values = self.adata.obs[col_name].values
        map_vals = dict([
            (x, ontology.convert_to_name(x)) if from_id else
            (x, ontology.convert_to_id(x))
            for x in np.unique([
                xx for xx in input_values
                if (xx.lower() not in map_exceptions and xx is not None)
            ])
        ])
        output_values = [
            map_vals[x] if x in map_vals.keys() else map_exceptions_value
            for x in input_values
        ]
        key_out = getattr(adata_ids, attr) if from_id else getattr(adata_ids, attr) + adata_ids.onto_id_suffix
        self.adata.obs[key_out] = output_values

    @property
    def citation(self):
        """
        Return all information necessary to cite data set.

        :return:
        """
        return [self.author, self.year, self.doi_journal]

    # Meta data handling code: Reading, writing and selected properties. Properties are either set in constructor
    # (and saved in self._somename) or accessed in self.meta.

    @property
    def meta_fn(self):
        if self.meta_path is None:
            meta = self.data_dir
        else:
            meta = os.path.join(self.meta_path, self.directory_formatted_doi)
        if meta is None:
            return None
        else:
            return os.path.join(meta, "meta", self.doi_cleaned_id + "_meta.csv")

    def load_meta(self, fn: Union[PathLike, str, None]):
        if fn is None:
            if self.meta_fn is not None:
                fn = self.meta_fn
        else:
            if isinstance(fn, str):
                fn = os.path.normpath(fn)
        # Only load meta data if file exists:
        if fn is not None and os.path.isfile(fn):
            meta = pandas.read_csv(
                fn,
                usecols=list(META_DATA_FIELDS.keys()),
            )
            # using dtype in read_csv through errors some times.
            for k, v in META_DATA_FIELDS.items():
                if k in meta.columns:
                    if meta[k].values[0] is not None:
                        meta[k] = np.asarray(meta[k].values, dtype=v)
            self.meta = meta.fillna("None").replace({"None": None})

    def write_meta(
            self,
            fn_meta: Union[None, str] = None,
            dir_out: Union[None, str] = None,
    ):
        """
        Write meta data object for data set.

        Does not cache data and attempts to load raw data.

        :param fn_meta: File to write to, selects automatically based on self.meta_path and self.id otherwise.
        :param dir_out: Path to write to, file name is selected automatically based on self.id.
        :return:
        """
        if fn_meta is not None and dir_out is not None:
            raise ValueError("supply either fn_meta or dir_out but not both")
        elif fn_meta is None and dir_out is None:
            if self.meta_fn is None:
                raise ValueError("provide either fn in load or via constructor (meta_path)")
            fn_meta = self.meta_fn
        elif fn_meta is None and dir_out is not None:
            fn_meta = os.path.join(dir_out, self.doi_cleaned_id + "_meta.csv")
        elif fn_meta is not None and dir_out is None:
            pass  # fn_meta is used
        else:
            assert False, "bug in switch"

        if self.adata is None:
            self.load(load_raw=True, allow_caching=False)
        # Add data-set wise meta data into table:
        meta = pandas.DataFrame(index=range(1))
        # Expand table by variably cell-wise or data set-wise meta data:
        for x in self._adata_ids.controlled_meta_fields:
            if hasattr(self, f"{x}_obs_key") and getattr(self, f"{x}_obs_key") is not None:
                col = getattr(self._adata_ids, x)
                meta[col] = (self.adata.obs[col].unique(), )
                if x in self._adata_ids.ontology_constrained:
                    col = getattr(self._adata_ids, x + self._adata_ids.onto_id_suffix)
                    meta[col] = (self.adata.obs[col].unique(), )
            else:
                meta[getattr(self._adata_ids, x)] = getattr(self, x)
        meta.to_csv(fn_meta)

    def set_dataset_id(
            self,
            idx: int = 1
    ):
        if self.sample_fn is not None:
            idx += self._sample_fns.index(self.sample_fn)
        idx = str(idx).zfill(3)

        if isinstance(self.author, List):
            author = self.author[0]
        else:
            author = self.author

        # Note: access private attributes here, e.g. _organism, to avoid loading of content via meta data, which would
        # invoke call to self.id before it is set.
        self.id = f"{clean_id_str(self._organism)}_" \
                  f"{clean_id_str(self._organ)}_" \
                  f"{self._year}_" \
                  f"{clean_id_str(self._assay_sc)}_" \
                  f"{clean_id_str(author)}_" \
                  f"{idx}_" \
                  f"{self.doi_main}"

    # Properties:

    @property
    def additional_annotation_key(self) -> Union[None, str]:
        return self._additional_annotation_key

    @additional_annotation_key.setter
    def additional_annotation_key(self, x: str):
        self._additional_annotation_key = x

    @property
    def annotated(self) -> Union[bool, None]:
        if self.cell_type_obs_key is not None:
            return True
        else:
            return False

    @property
    def assay_sc(self) -> Union[None, str]:
        return self._assay_sc

    @assay_sc.setter
    def assay_sc(self, x: str):
        x = self._value_protection(attr="assay_sc", allowed=self.get_ontology(k="assay_sc"), attempted=x)
        self._assay_sc = x

    @property
    def assay_differentiation(self) -> Union[None, str]:
        return self._assay_differentiation

    @assay_differentiation.setter
    def assay_differentiation(self, x: str):
        x = self._value_protection(attr="assay_differentiation",
                                   allowed=self.get_ontology(k="assay_differentiation"), attempted=x)
        self._assay_differentiation = x

    @property
    def assay_type_differentiation(self) -> Union[None, str]:
        return self._assay_type_differentiation

    @assay_type_differentiation.setter
    def assay_type_differentiation(self, x: str):
        x = self._value_protection(attr="assay_type_differentiation",
                                   allowed=self.get_ontology(k="assay_type_differentiation"), attempted=x)
        self._assay_type_differentiation = x

    @property
    def bio_sample(self) -> Union[None, str]:
        if self._bio_sample is not None and self._bio_sample != self._adata_ids.unknown_metadata_identifier:
            return self._bio_sample
        else:
            # Define biological sample automatically.
            bio_sample = "*".join([x for x in [
                self.assay_sc,
                self.assay_differentiation,
                self.assay_type_differentiation,
                self.development_stage,
                self.disease,
                self.ethnicity,
                self.individual,
                self.organ,
                self.organism,
                self.sex,
            ] if x is not None])
            return bio_sample if bio_sample != "" else None

    @bio_sample.setter
    def bio_sample(self, x: str):
        self._bio_sample = x

    @property
    def bio_sample_obs_key(self) -> Union[None, str]:
        if self._bio_sample_obs_key is not None and \
                self._bio_sample_obs_key != self._adata_ids.unknown_metadata_identifier:
            return self._bio_sample_obs_key
        else:
            # Define biological sample automatically.
            bio_sample_obs_key = "*".join([x for x in [
                self.assay_sc_obs_key,
                self.assay_differentiation_obs_key,
                self.assay_type_differentiation_obs_key,
                self.development_stage_obs_key,
                self.disease_obs_key,
                self.ethnicity_obs_key,
                self.individual_obs_key,
                self.organ_obs_key,
                self.organism_obs_key,
                self.sex_obs_key,
                self.state_exact_obs_key,
            ] if x is not None])
            return bio_sample_obs_key if bio_sample_obs_key != "" else None

    @bio_sample_obs_key.setter
    def bio_sample_obs_key(self, x: str):
        self._bio_sample_obs_key = x

    @property
    def cell_line(self) -> Union[None, str]:
        return self._cell_line

    @cell_line.setter
    def cell_line(self, x: str):
        self._cell_line = x

    @property
    def cell_type(self) -> Union[None, str]:
        return self._cell_type

    @cell_type.setter
    def cell_type(self, x: str):
        self._cell_type = x

    @property
    def data_dir(self):
        # Data is either directly in user supplied directory or in a sub directory if the overall directory is managed
        # by sfaira: In this case, the sub directory is named after the doi of the data set.
        if self.data_dir_base is None:
            return None
        else:
            sfaira_path = os.path.join(self.data_dir_base, self.directory_formatted_doi)
            # Allow checking in secondary path, named after second DOI associated with study.
            # This allows association of raw data already downloaded even after DOI is updated.
            if self.doi_preprint is not None:
                sfaira_path_secondary = os.path.join(self.data_dir_base,
                                                     get_directory_formatted_doi(x=self.doi_preprint))
            else:
                sfaira_path_secondary = None
            if os.path.exists(sfaira_path):
                return sfaira_path
            elif self.doi_preprint is not None and os.path.exists(sfaira_path_secondary):
                return sfaira_path_secondary
            else:
                return self.data_dir_base

    @property
    def default_embedding(self) -> Union[None, str]:
        return self._default_embedding

    @default_embedding.setter
    def default_embedding(self, x: str):
        x = self._value_protection(attr="default_embedding", allowed=self.get_ontology(k="default_embedding"),
                                   attempted=x)
        self._default_embedding = x

    @property
    def development_stage(self) -> Union[None, str]:
        return self._development_stage

    @development_stage.setter
    def development_stage(self, x: str):
        x = self._value_protection(attr="development_stage",
                                   allowed=self.get_ontology(k="development_stage"),
                                   attempted=x)
        self._development_stage = x

    @property
    def disease(self) -> Union[None, str]:
        return self._disease

    @disease.setter
    def disease(self, x: str):
        x = self._value_protection(attr="disease", allowed=self.get_ontology(k="disease"),
                                   attempted=x)
        self._disease = x

    @property
    def doi_journal(self) -> str:
        """
        The prepring publication (secondary) DOI associated with the study.
        See also `.doi_journal`.
        """
        return self._doi_journal

    @doi_journal.setter
    def doi_journal(self, x: str):
        self._doi_journal = x

    @property
    def doi_preprint(self) -> str:
        """
        The journal publication (main) DOI associated with the study.
        See also `.doi_preprint`.
        """
        return self._doi_preprint

    @doi_preprint.setter
    def doi_preprint(self, x: str):
        self._doi_preprint = x

    @property
    def doi(self) -> List[str]:
        """
        All publication DOI associated with the study which are the journal publication and the preprint.
        See also `.doi_preprint`, `.doi_journal`.
        """
        dois = []
        if self.doi_journal is not None:
            dois.append(self.doi_journal)
        if self.doi_preprint is not None:
            dois.append(self.doi_preprint)
        return dois

    @property
    def doi_main(self) -> str:
        """
        The main DOI associated with the study which is the journal publication if available, otherwise the preprint.
        See also `.doi_preprint`, `.doi_journal`.
        """
        return self.doi_preprint if self.doi_journal is None else self.doi_journal

    @property
    def directory_formatted_doi(self) -> str:
        return get_directory_formatted_doi(x=self.doi_main)

    @property
    def download_url_data(self) -> Union[Tuple[List[str]], Tuple[List[None]]]:
        """
        Data download website(s).

        Save as tuple with single element, which is a list of all download websites relevant to dataset.
        :return:
        """
        x = self._download_url_data
        if isinstance(x, str) or x is None:
            x = [x]
        if isinstance(x, list):
            x = (x,)
        return x

    @download_url_data.setter
    def download_url_data(self, x: Union[str, None, List[str], Tuple[str], List[None], Tuple[None]]):
        # Formats to tuple with single element, which is a list of all download websites relevant to dataset,
        # which can be used as a single element column in a pandas data frame.
        if isinstance(x, str) or x is None:
            x = [x]
        if isinstance(x, list):
            x = (x,)
        self._download_url_data = x

    @property
    def download_url_meta(self) -> Union[Tuple[List[str]], Tuple[List[None]]]:
        """
        Meta data download website(s).

        Save as tuple with single element, which is a list of all download websites relevant to dataset.
        :return:
        """
        x = self._download_url_meta
        # if self._download_url_meta is not None:  # TODO add this back in once download_meta is set in all datasets
        #    x = self._download_url_meta
        # else:
        #    if self.meta is None:
        #        self.load_meta(fn=None)
        #    x = self.meta[self._adata_ids.download_url_meta]
        if isinstance(x, str) or x is None:
            x = [x]
        if isinstance(x, list):
            x = (x,)
        return x

    @download_url_meta.setter
    def download_url_meta(self, x: Union[str, None, List[str], Tuple[str], List[None], Tuple[None]]):
        # Formats to tuple with single element, which is a list of all download websites relevant to dataset,
        # which can be used as a single element column in a pandas data frame.
        if isinstance(x, str) or x is None:
            x = [x]
        if isinstance(x, list):
            x = (x,)
        self._download_url_meta = x

    @property
    def ethnicity(self) -> Union[None, str]:
        return self._ethnicity

    @ethnicity.setter
    def ethnicity(self, x: str):
        x = self._value_protection(attr="ethnicity", allowed=self.get_ontology(k="ethnicity"),
                                   attempted=x)
        self._ethnicity = x

    @property
    def feature_reference(self) -> str:
        return self._feature_reference

    @feature_reference.setter
    def feature_reference(self, x: str):
        self._feature_reference = x

    @property
    def feature_type(self) -> str:
        return self._feature_type

    @feature_type.setter
    def feature_type(self, x: str):
        x = self._value_protection(attr="feature_type", allowed=self.get_ontology(k="feature_type"),
                                   attempted=x)
        self._feature_type = x

    @property
    def id(self) -> str:
        if self._id is not None:
            return self._id
        else:
            raise AttributeError(f"Dataset ID was not set in dataloader in {self.doi_main}, please ensure the "
                                 f"dataloader constructor of this dataset contains a call to self.set_dataset_id()")

    @id.setter
    def id(self, x: str):
        self._id = x

    @property
    def individual(self) -> Union[None, str]:
        return self._individual

    @individual.setter
    def individual(self, x: str):
        self._individual = x

    @property
    def loaded(self) -> bool:
        """
        :return: Whether DataSet was loaded into memory.
        """
        return self.adata is not None

    @property
    def meta(self) -> Union[None, pd.DataFrame]:
        return self._meta

    @meta.setter
    def meta(self, x: Union[None, pd.DataFrame]):
        # Make sure formatting is correct:
        if x is not None:
            for k, v in x.items():
                v = v.tolist()  # avoid numpy data types
                if k not in META_DATA_FIELDS.keys():
                    raise ValueError(f"did not find {k} in format look up table")
                else:
                    if x[k] is not None:  # None is always allowed.
                        if not isinstance(v[0], META_DATA_FIELDS[k]):
                            raise ValueError(f"key '{k}' of value `{v[0]}` and signature `{type(v[0])}` "
                                             f"in meta data table did not match signature "
                                             f"{str(META_DATA_FIELDS[k])}")
        self._meta = x

    @property
    def ncells(self) -> Union[None, int]:
        # ToDo cache this if it was loaded from meta?
        if self.adata is not None:
            return self.adata.n_obs
        else:
            return None

    @property
    def primary_data(self) -> Union[None, bool]:
        return self._primary_data

    @primary_data.setter
    def primary_data(self, x: bool):
        x = self._value_protection(attr="primary_data", allowed=self.get_ontology(k="primary_data"),
                                   attempted=x)
        self._primary_data = x

    @property
    def organ(self) -> Union[None, str]:
        return self._organ

    @organ.setter
    def organ(self, x: str):
        x = self._value_protection(attr="organ", allowed=self.get_ontology(k="organ"), attempted=x)
        self._organ = x

    @property
    def organism(self) -> Union[None, str]:
        return self._organism

    @organism.setter
    def organism(self, x: str):
        x = self._value_protection(attr="organism", allowed=self.get_ontology(k="organism"), attempted=x)
        # Update ontology container so that correct ontologies are queried:
        self.ontology_container_sfaira.organism_cache = x
        self._organism = x

    @property
    def sample_source(self) -> Union[None, str]:
        return self._sample_source

    @sample_source.setter
    def sample_source(self, x: str):
        x = self._value_protection(attr="sample_source", allowed=self.get_ontology(k="sample_source"),
                                   attempted=x)
        self._sample_source = x

    @property
    def sex(self) -> Union[None, str]:
        return self._sex

    @sex.setter
    def sex(self, x: str):
        x = self._value_protection(attr="sex", allowed=self.get_ontology(k="sex"), attempted=x)
        self._sex = x

    @property
    def source(self) -> str:
        return self._source

    @source.setter
    def source(self, x: Union[str, None]):
        self._source = x

    @property
    def state_exact(self) -> Union[None, str]:
        return self._state_exact

    @state_exact.setter
    def state_exact(self, x: str):
        self._state_exact = x

    @property
    def tech_sample(self) -> Union[None, str]:
        if self._tech_sample is not None and self._tech_sample != self._adata_ids.unknown_metadata_identifier:
            return self._tech_sample
        else:
            # Define technical batch automatically as biological sample.
            return self.bio_sample

    @tech_sample.setter
    def tech_sample(self, x: str):
        self._tech_sample = x

    @property
    def tech_sample_obs_key(self) -> Union[None, str]:
        if self._tech_sample_obs_key is not None and \
                self._tech_sample_obs_key != self._adata_ids.unknown_metadata_identifier:
            return self._tech_sample_obs_key
        else:
            # Define technical batch automatically as biological sample.
            return self.bio_sample_obs_key

    @tech_sample_obs_key.setter
    def tech_sample_obs_key(self, x: str):
        self._tech_sample_obs_key = x

    @property
    def author(self) -> str:
        return self._author

    @author.setter
    def author(self, x: str):
        self._author = x

    @property
    def title(self):
        if self._title is not None:
            return self._title
        else:
            return self.__crossref_query(k="title")

    @property
    def year(self) -> Union[None, int]:
        return self._year

    @year.setter
    def year(self, x: int):
        x = self._value_protection(attr="year", allowed=self.get_ontology(k="year"), attempted=x)
        self._year = x

    @property
    def celltypes_universe(self):
        if self._celltype_universe is None:
            self._celltype_universe = CelltypeUniverse(
                cl=getattr(self.ontology_container_sfaira, "cell_type"),
                uberon=getattr(self.ontology_container_sfaira, "organ"),
            )
        return self._celltype_universe

    @property
    def ontology_class_maps(self) -> Dict[str, pd.DataFrame]:
        return self._ontology_class_maps

    @ontology_class_maps.setter
    def ontology_class_maps(self, x: Dict[str, pd.DataFrame]):
        for k, v in x.items():
            assert v.shape[1] in [2, 3], f"{v.shape} in {self.id}"
            assert v.columns[0] == self._adata_ids.classmap_source_key
            assert v.columns[1] == self._adata_ids.classmap_target_key
            # Check for weird entries:
            # nan arises if columns was empty in that row.
            nan_vals = np.where([
                False if isinstance(x, str) else (np.isnan(x) or x is None)
                for x in v[self._adata_ids.classmap_target_key].values.tolist()
            ])[0]
            assert len(nan_vals) == 0, \
                f"Found nan target values in {self.id} for {x[self._adata_ids.classmap_target_key].values[nan_vals]}," \
                f" check if all entries in cell type .tsv file are non-empty, for example." \
                f"This bug may arise if a tab separator of columns is missing in one or multiple rows, for example."
            # Transform data frame into a mapping dictionary:
            self._ontology_class_maps[k] = dict(list(zip(
                v[self._adata_ids.classmap_source_key].values.tolist(),
                v[self._adata_ids.classmap_target_key].values.tolist()
            )))

    def __crossref_query(self, k):
        """
        Queries cross REST API via package crossref_commons.

        :param k: Key to extract from crossref query container.
        :return:
        """
        from crossref_commons.retrieval import get_entity
        from crossref_commons.types import EntityType, OutputType
        try:
            attempt_counter = 0
            while True:
                x = None
                try:
                    attempt_counter += 1
                    x = get_entity(self.doi_main, EntityType.PUBLICATION, OutputType.JSON)[k]
                except ConnectionError as e:
                    # Terminate trial after 5 attempts with ConnectionError:
                    if attempt_counter > 5:
                        raise ConnectionError(e)
                finally:
                    if k == "author":
                        pass
                    return x
        except ValueError as e:
            print(f"ValueError: {e}")
            return None
        except ConnectionError as e:
            print(f"ConnectionError: {e}")
            return None

    def _value_protection(
            self,
            attr: str,
            allowed: Union[Ontology, None],
            attempted
    ):
        """
        Check whether value is from set of allowed values.

        Does not check if allowed is None.
        Cleans entry to term name if ontology ID is provided.

        :param attr: Attribute to set.
        :param allowed: Constraint for values of `attr`.
            Either ontology instance used to constrain entries, or list of allowed values.
        :param attempted: Value(s) to attempt to set in `attr`.
        :return:
        """
        if not isinstance(attempted, list):
            if isinstance(attempted, np.ndarray):
                attempted_ls = attempted.tolist()
            elif isinstance(attempted, tuple):
                attempted_ls = list(attempted)
            else:
                attempted_ls = [attempted]
        else:
            attempted_ls = attempted
        attempted_clean = []
        for x in attempted_ls:
            if allowed is None:
                attempted_clean.append(x)
            elif isinstance(allowed, Ontology):
                if attr == "disease" and (x.lower() == "normal" or x.lower() == "healthy"):
                    # TODO required because of missing streamlining between sfaira and 10x, remove in future.
                    attempted_clean.append("healthy")
                elif x in allowed.node_names:
                    attempted_clean.append(x)
                else:
                    if isinstance(allowed, OntologyHierarchical) and x in allowed.node_ids:
                        attempted_clean.append(allowed.convert_to_name(x))
                    else:
                        raise ValueError(f"'{x}' is not a valid entry for {attr} in data set {self.doi}.")
            else:
                raise ValueError(f"argument allowed of type {type(allowed)} is not a valid entry for {attr}.")
        # Flatten attempts if only one was made:
        if len(attempted_clean) == 1:
            attempted_clean = attempted_clean[0]
        return attempted_clean

    def subset_cells(self, key, values):
        """
        Subset list of adata objects based on cell-wise properties.

        These keys are properties that are not available in lazy model and require loading first because the
        subsetting works on the cell-level: .adata are maintained but reduced to matches.

        :param key: Property to subset by. Options:

            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to.
        :return:
        """
        if not isinstance(values, list):
            values = [values]

        def get_subset_idx(samplewise_key, cellwise_key):
            try:
                sample_attr = getattr(self, samplewise_key)
                if not isinstance(sample_attr, list):
                    sample_attr = [sample_attr]
            except AttributeError:
                sample_attr = None
            obs_key = getattr(self, cellwise_key)
            if sample_attr is not None and len(sample_attr) == 1:
                # Only use sample-wise subsetting if the sample-wise attribute is unique (not mixed).
                if np.any([x in values for x in sample_attr]):
                    idx = np.arange(1, self.ncells)
                else:
                    idx = np.array([])
            elif obs_key is not None:
                assert self.adata is not None, "adata was not yet loaded"
                values_found = self.adata.obs[obs_key].values
                values_found_unique = np.unique(values_found)
                try:
                    ontology = getattr(self.ontology_container_sfaira, samplewise_key)
                except AttributeError:
                    raise ValueError(f"{key} not a valid property of ontology_container object")
                # Test only unique elements  found in ontology to save time.
                values_found_unique_matched = [
                    x for x in values_found_unique if np.any([
                        is_child(query=x, ontology=ontology, ontology_parent=y)
                        for y in values
                    ])
                ]
                # TODO keep this logging for now to catch undesired behaviour resulting from loaded edges in ontologies.
                print(f"matched cell-wise keys {str(values_found_unique_matched)} in data set {self.id}")
                idx = np.where([x in values_found_unique_matched for x in values_found])[0]
            else:
                assert False, "no subset chosen"
            return idx

        idx_keep = get_subset_idx(samplewise_key=key, cellwise_key=key + "_obs_key")
        self.adata = self.adata[idx_keep, :].copy() if len(idx_keep) > 0 else None

    def show_summary(self):
        print(f"{(self.supplier, self.organism, self.organ, self.assay_sc, self.disease)}")

    def __assert_loaded(self):
        if self.adata is None:
            raise ValueError("adata was not loaded, this is necessary for this operation")



"""
Helper functionalities to interact with cellxgene REST API.
"""

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

CELLXGENE_PRODUCTION_ENDPOINT = 'https://api.cellxgene.cziscience.com'
COLLECTIONS = "/dp/v1/collections/"
DOWNLOAD_DATASET = "/dp/v1/datasets/"
MAX_RETRIES = 5
TIMEOUT_COLLECTION = 5
TIMEOUT_DATA = 10
HTTP_ERROR_LIST = [429, 502, 504]


class CustomHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout, **kwargs):
        self.timeout = timeout
        super().__init__(**kwargs)

    def send(self, request, **kwargs):
        kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def rest_api_collection_request(url):
    retry_strategy = Retry(
        backoff_factor=0,
        method_whitelist=["GET"],
        status_forcelist=HTTP_ERROR_LIST,
        total=MAX_RETRIES,
    )
    adapter = CustomHTTPAdapter(timeout=TIMEOUT_COLLECTION, max_retries=retry_strategy)
    https = requests.Session()
    https.mount("https://", adapter)
    r = https.get(url)
    r.raise_for_status()
    return r.json()


def rest_api_data_request(presigned_url):
    retry_strategy = Retry(
        backoff_factor=0,
        method_whitelist=["GET"],
        status_forcelist=HTTP_ERROR_LIST,
        total=MAX_RETRIES,
    )
    adapter = CustomHTTPAdapter(timeout=TIMEOUT_DATA, max_retries=retry_strategy)
    https = requests.Session()
    https.mount("https://", adapter)
    r = https.get(presigned_url)
    return r.content


def get_collection(collection_id):
    return rest_api_collection_request(url=CELLXGENE_PRODUCTION_ENDPOINT + COLLECTIONS + collection_id)


def get_collections():
    return rest_api_collection_request(url=CELLXGENE_PRODUCTION_ENDPOINT + COLLECTIONS)['collections']


def get_data(presigned_url):
    return rest_api_data_request(presigned_url=presigned_url)



import anndata
import dask.array
import h5py
import numpy as np
import os
import pytest
import scipy.sparse

from sfaira.data import load_store

from sfaira.unit_tests.data_for_tests.loaders import PrepareData


@pytest.mark.parametrize("store_format", ["h5ad", "dao", "anndata"])
def test_fatal(store_format: str):
    """
    Test if basic methods of stores abort.
    """
    if store_format == "anndata":
        stores = PrepareData().prepare_store_anndata()
    else:
        store_path = PrepareData().prepare_store(store_format=store_format)
        stores = load_store(cache_path=store_path, store_format=store_format)
    stores.subset(attr_key="organism", values=["Mus musculus"])
    store = stores.stores["Mus musculus"]
    # Test both single and multi-store:
    for x in [store, stores]:
        _ = x.n_obs
        _ = x.n_vars
        _ = x.var_names
        _ = x.shape
        _ = x.indices
        _ = x.genome_container


@pytest.mark.parametrize("store_format", ["h5ad", "dao"])
def test_config(store_format: str):
    """
    Test that data set config files can be set, written and recovered.
    """
    store_path = PrepareData().prepare_store(store_format=store_format)
    config_path = os.path.join(store_path, "config_lung")
    store = load_store(cache_path=store_path, store_format=store_format)
    store.subset(attr_key="organism", values=["Mus musculus"])
    store.subset(attr_key="assay_sc", values=["10x technology"])
    store.write_config(fn=config_path)
    store2 = load_store(cache_path=store_path, store_format=store_format)
    store2.load_config(fn=config_path + ".pickle")
    assert np.all(store.indices.keys() == store2.indices.keys())
    assert np.all([np.all(store.indices[k] == store2.indices[k])
                   for k in store.indices.keys()])


@pytest.mark.parametrize("store_format", ["h5ad", "dao"])
def test_store_data(store_format: str):
    """
    Test if the data exposed by the store are the same as in the original Dataset instance after streamlining.
    """
    data = PrepareData()
    # Run standard streamlining workflow on dsg and compare to object relayed via store.
    # Prepare dsg.
    dsg = data.prepare_dsg(load=True)
    # Prepare store.
    # Rewriting store to avoid mismatch of randomly generated data in cache and store.
    store_path = data.prepare_store(store_format=store_format, rewrite=False, rewrite_store=True)
    store = load_store(cache_path=store_path, store_format=store_format)
    store.subset(attr_key="doi_journal", values=["no_doi_mock1"])
    dataset_id = store.adata_by_key[list(store.indices.keys())[0]].uns["id"]
    adata_store = store.adata_by_key[dataset_id]
    x_store = store.data_by_key[dataset_id]
    adata_ds = dsg.datasets[dataset_id].adata
    x_ds = adata_ds.X.todense()
    if isinstance(x_store, dask.array.Array):
        x_store = x_store.compute()
    if isinstance(x_store, h5py.Dataset):
        # Need to load sparse matrix into memory if it comes from a backed anndata object.
        x_store = x_store[:, :]
    if isinstance(x_store, anndata._core.sparse_dataset.SparseDataset):
        # Need to load sparse matrix into memory if it comes from a backed anndata object.
        x_store = x_store[:, :]
    if isinstance(x_store, scipy.sparse.csr_matrix):
        x_store = x_store.todense()
    if isinstance(x_ds, anndata._core.sparse_dataset.SparseDataset):
        # Need to load sparse matrix into memory if it comes from a backed anndata object.
        x_ds = x_ds[:, :]
    if isinstance(x_ds, scipy.sparse.csr_matrix):
        x_ds = x_ds.todense()
    # Check that non-zero elements are the same:
    assert x_store.shape[0] == x_ds.shape[0]
    assert x_store.shape[1] == x_ds.shape[1]
    assert np.all(np.where(x_store > 0)[0] == np.where(x_ds > 0)[0]), (np.sum(x_store > 0), np.sum(x_ds > 0))
    assert np.all(np.where(x_store > 0)[1] == np.where(x_ds > 0)[1]), (np.sum(x_store > 0), np.sum(x_ds > 0))
    assert np.all(x_store - x_ds == 0.), (np.sum(x_store), np.sum(x_ds))
    assert x_store.dtype == x_ds.dtype
    # Note: Do not run test on sum across entire object if dtype is float32 as this can result in test failures because
    # of float overflows.
    # Check .obs
    obs_store = adata_store.obs
    obs_ds = adata_ds.obs
    assert np.all(obs_store.columns == obs_ds.columns), (obs_store.columns, obs_ds.columns)
    for k, v in obs_store.items():
        assert np.all(np.asarray(v.values.tolist()) == np.asarray(obs_ds[k].values.tolist()))
    # Check .var
    var_store = adata_store.var
    var_ds = adata_ds.var
    assert np.all(var_store.columns == var_ds.columns), (var_store.columns, var_ds.columns)
    for k, v in var_store.items():
        assert np.all(np.asarray(v.values.tolist()) == np.asarray(var_ds[k].values.tolist()))
    # Check .uns
    uns_store = adata_store.uns
    uns_ds = adata_ds.uns
    assert np.all(uns_store.keys() == uns_ds.keys()), (uns_store.keys(), uns_ds.keys())
    for k, v in uns_store.items():
        assert np.all(v == uns_ds[k])



import os
import pytest
import shutil
from typing import List

from sfaira.unit_tests.directories import DIR_DATA_DATABASES_CACHE
from sfaira.unit_tests.data_for_tests.databases.utils import prepare_dsg_database
from sfaira.unit_tests.data_for_tests.databases.consts import CELLXGENE_DATASET_ID


# Execute this one first so that data sets are only downloaded once. Named test_a for this reason.
@pytest.mark.parametrize("database", [
    ("cellxgene", ["id", CELLXGENE_DATASET_ID]),
])
def test_a_dsgs_download(database: str):
    """
    Tests if downloading of data base entries works.

    Warning, deletes entire database unit test cache.
    """
    database, subset_args = database
    if os.path.exists(DIR_DATA_DATABASES_CACHE):
        shutil.rmtree(DIR_DATA_DATABASES_CACHE)
    dsg = prepare_dsg_database(database=database, download=False)
    if subset_args is not None:
        dsg.subset(key=subset_args[0], values=subset_args[1])
    dsg.download()


@pytest.mark.parametrize("database", [
    ("cellxgene", ["id", CELLXGENE_DATASET_ID]),
    ("cellxgene", ["organism", "Homo sapiens"]),
])
def test_dsgs_subset(database: str):
    """
    Tests if subsetting results only in datasets of the desired characteristics.
    """
    database, subset_args = database
    dsg = prepare_dsg_database(database=database)
    dsg.subset(key=subset_args[0], values=subset_args[1])



try:
    import tensorflow as tf
except ImportError:
    tf = None
import numpy as np


class WarmUpTraining(tf.keras.callbacks.Callback):
    def __init__(
            self,
            fn=None,
            max_beta=1.0,
            verbose=1
    ):
        super(WarmUpTraining, self).__init__()
        self.verbose = verbose
        self.beta = None
        self.fn = fn
        self.count = 0
        self.max_beta = max_beta

    def on_epoch_begin(self, epoch, logs=None):
        if epoch == 0:
            if not hasattr(self.model.loss["kl"], 'beta'):
                raise ValueError('Model must have a "beta" attribute.')

        self.beta = np.minimum(self.max_beta, (epoch + 1) / 100)   # sigmoid((x-125)/4)
        tf.keras.backend.set_value(self.model.loss["kl"].beta, self.beta)
        if self.verbose > 0:
            print('\nSet beta to %s' % self.beta)

        if self.fn is not None and epoch == 99:
            self.model.save_weights(self.fn + "_weights_wup.h5")
            if self.verbose > 0:
                print('\nSave weights after warm-up training at epoch %03d' % (epoch + 1))


class LearningRateSchedule(tf.keras.callbacks.Callback):
    def __init__(
            self,
            verbose=1
    ):
        super(LearningRateSchedule, self).__init__()
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs=None):
        # pseudo_inputs = self.model.layers[3].pseudo_inputs_layer.pseudo_inputs.numpy()
        # print("\nepoch %s pseudo inputs: %s / %s / %s  (max / mean / min) " %
        #       (epoch + 1, pseudo_inputs.max(), pseudo_inputs.mean(), pseudo_inputs.min()))
        if epoch == 199:
            lr = tf.keras.backend.get_value(self.model.optimizer.lr)
            tf.keras.backend.set_value(self.model.optimizer.lr, lr / 10)
            if self.verbose > 0:
                print('\nReduce lr training at epoch %03d to %s' % (epoch + 1, lr / 10))

        if epoch == 249:
            lr = tf.keras.backend.get_value(self.model.optimizer.lr)
            tf.keras.backend.set_value(self.model.optimizer.lr, lr / 10)
            if self.verbose > 0:
                print('\nReduce lr training at epoch %03d to %s' % (epoch + 1, lr / 10))

        if epoch == 299:
            self.model.stop_training = True
            if self.verbose > 0:
                print('\nTerminate training at epoch %03d' % (epoch + 1))



import os
import pandas as pd
import anndata as ad
import scipy.sparse
import numpy as np


def load(data_dir, sample_fn, **kwargs):
    fn = [
        os.path.join(data_dir, f"GSE114374_Human_{sample_fn}_expression_matrix.txt.gz"),
        os.path.join(data_dir, f"{sample_fn.lower()}_meta_data_stromal_with_donor.txt"),
    ]
    matrix = pd.read_csv(fn[0], sep="\t")
    obs = pd.read_csv(fn[1], sep="\t", index_col=3)
    adata = ad.AnnData(matrix.T)
    adata.X = scipy.sparse.csc_matrix(np.expm1(adata.X))
    adata.obs = obs
    s_dict = {"F": "female", "M": "male"}
    adata.obs['Sex'] = [s_dict[i] for i in adata.obs['Sex']]
    adata.obs['Age'] = [str(x) + "-year-old homosapiens stage" for x in adata.obs['Age'].values]

    return adata



import anndata
import numpy as np
import pandas as pd
import pytest
import scipy.sparse
from typing import Union

from sfaira.data.utils import map_freetext_to_ontology, collapse_matrix


@pytest.mark.parametrize("trial_cell_type_labels",
                         ["type B pancreatic cell", "beta", ["type B pancreatic cell", "beta"]])
@pytest.mark.parametrize("choices_for_perfect_match", [True, False])
@pytest.mark.parametrize("anatomical_constraint", [None, "pancreas"])
def test_map_celltype_to_ontology(
        trial_cell_type_labels: str,
        choices_for_perfect_match: bool,
        anatomical_constraint: Union[str, None]
):
    trial_cell_type_labels = [trial_cell_type_labels] if isinstance(trial_cell_type_labels, str) \
        else trial_cell_type_labels
    perfectly_matched_query = ["type B pancreatic cell" == x for x in trial_cell_type_labels]
    matches = map_freetext_to_ontology(
        queries=trial_cell_type_labels,
        onto="cell_type",
        organism="Homo sapiens",
        include_synonyms=True,
        anatomical_constraint=anatomical_constraint,
        choices_for_perfect_match=choices_for_perfect_match,
        always_return_dict=False,
    )
    for x, y in zip(trial_cell_type_labels, perfectly_matched_query):
        if isinstance(matches, dict):  # dictionary over queries with list of matches as value each
            if y and not choices_for_perfect_match:
                assert isinstance(matches[x], str), matches
                assert matches[x] == "type B pancreatic cell"
            else:
                assert isinstance(matches[x], list), matches
                assert "type B pancreatic cell" in matches[x]
        else:  # matches for single query
            if y and not choices_for_perfect_match:
                assert isinstance(matches, str), matches
                assert matches == "type B pancreatic cell"
            else:
                assert isinstance(matches, list), matches
                assert "type B pancreatic cell" in matches


@pytest.mark.parametrize("data", ["scipy.sparse.lil_matrix", "scipy.sparse.csr_matrix", "numpy"])
@pytest.mark.parametrize("duplications", [False, True])
def test_collapse_matrix(
        data: str,
        duplications: bool,
):
    """
    Tests collapse_matrix.

    Tests if:

        - matrix type is maintained
        - row sums are maintained
        - feature dimension is correct

    :param data: Data format.
    :param duplications: Whether feature names are duplicated.
    :return:
    """
    x = np.asarray(np.random.randint(0, 100, size=(10, 10)), dtype="float32")
    if data == "scipy.sparse.lil_matrix":
        x = scipy.sparse.lil_matrix(x)
    elif data == "scipy.sparse.csr_matrix":
        x = scipy.sparse.csr_matrix(x)
    elif data == "numpy":
        pass
    else:
        assert False
    if duplications:
        # Create triplicate and duplicate gene names:
        index = ["g" + str(i) for i in range(2)] + ["g" + str(i) for i in range(3)] + \
                ["g" + str(i) for i in range(x.shape[1] - 3 - 2)]
    else:
        index = ["g" + str(i) for i in range(x.shape[1])]
    adata = anndata.AnnData(x, var=pd.DataFrame({"var_column": index}))
    adata2 = collapse_matrix(adata=adata, var_column="var_column")
    assert adata.X.shape[0] == adata2.X.shape[0], "observation dimension mismatch"
    assert adata.X.dtype == adata2.X.dtype, "type mismatch"
    assert adata2.X.shape[1] == len(np.unique(adata.var["var_column"])), "feature dimension mismatch"
    assert np.all(np.asarray(adata.X.sum()).flatten() == np.asarray(adata2.X.sum().flatten())), \
        "total count mismatch"
    assert np.all(np.asarray(adata.X.sum(axis=1)).flatten() == np.asarray(adata2.X.sum(axis=1).flatten())), \
        "observation-wise count mismatch"



import anndata
import os
import scipy.io
import tarfile
import gzip
import pandas as pd


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, "GSE143317_RAW.tar")

    with tarfile.open(fn) as archive:
        file = f'{sample_fn}_filtered_feature_bc_matrix.tar.gz'
        with tarfile.open(fileobj=archive.extractfile(file)) as sample:
            name = sample.getnames()[0]
            with gzip.open(sample.extractfile(name + '/matrix.mtx.gz'), "rb") as mm:
                x = scipy.io.mmread(mm).T.tocsr()
            obs = pd.read_csv(sample.extractfile(name + "/barcodes.tsv.gz"), compression="gzip",
                              header=None, sep="\t", index_col=0)
            obs.index.name = None
            var = pd.read_csv(sample.extractfile(name + "/features.tsv.gz"), compression="gzip",
                              header=None, sep="\t")
            var.columns = ["ensembl", "names", "feature_types"]
            var.index = var["ensembl"].values
            adata = anndata.AnnData(X=x, obs=obs, var=var)
    return adata
